-- Affine Function Library
-- Gordon Pace - Gerardo Schneider - VERIMAG 
-- 03/07/2001

module AffineFunctions where

import Infinity
import Interval

-- Generic classes to overload the application of a function (apply)
-- composition (comp), compsition of a list of functions (mcomp) 
-- and inversion (invert)
class Applicable a b where
  apply  :: a -> b -> b

class Function a where
  forallInvert :: a -> a
  invert :: a -> a

  comp :: a -> a -> a

  mcomp :: [a] -> a
  mcomp = foldl1 comp
  -- mcomp [x] = x
  -- mcomp (x:xs) = x `comp` (mcomp xs)
  -- given [fn, ..., f2, f1] it gives fn o .... o f2 o f1

  funcId :: a


-- ============================================
-- Affine functions
-- ============================================
-- an Af is defined as af(x) = a*x+b

newtype AF = AF (RealPlus, RealPlus)
  deriving Eq

instance Show AF where
  show (AF (m,c)) 
    | m==0 && c==0 = "0"
    | c==0         = show_m
    | m==0         = show c
    | c<0          = show_m ++ show c
    | otherwise    = show_m ++"+"++show c
    where
      show_m 
        | m == 1 = "x"
        | otherwise = show m ++ "x"
     
toAF :: (RealPlus, RealPlus) -> AF
toAF = AF

fromAF :: AF -> (RealPlus, RealPlus)
fromAF (AF x) = x


instance Applicable AF RealPlus where
  apply af = 
    let (a,b) = fromAF af 
    in  \x -> a * x + b -- \lambda x .(a * x + b)

instance Function AF where
-- invert (ax+b) = (x-b)/a
  invert af = 
    let (a,b) = fromAF af 
    in  toAF (1/a, -b/a)

  forallInvert = invert

-- comp af1 af2 = af2 o af1
  comp af1 af2 = 
    let (a1,b1) = fromAF af1
        (a2,b2) = fromAF af2  
    in  toAF (a1*a2, b1*a2 + b2)

-- Identity AF
  funcId = toAF (1,0)


-- ============================================
-- Multi-Valued Affine functions
-- ============================================
-- a MAf is a pair of AF: maf(<l,u>) = <a1*l+b1, a2*u+b2>

newtype MAF = MAF (AF, AF)
  deriving Eq

instance Show MAF where
  show (MAF (a1,a2)) = 
    "["++show a1++", "++show a2++"]"

toMAF :: (AF, AF) -> MAF
toMAF = MAF

fromMAF :: MAF -> (AF, AF)
fromMAF (MAF x) = x


instance Applicable MAF (Interval RealPlus) where
  apply maf interval
    | isEmpty interval  = emptyInt
    | otherwise         = 
        let (l,u) = fromInterval interval
            (af1,af2) = fromMAF maf 
            l_val = value l
            u_val = value u
            l_ext = clop l
            u_ext = clop u
        in toInterval (l_ext (apply af1 l_val), u_ext (apply af2 u_val))

instance Function MAF where
-- invert <f1,f2> = <invert f2, invert f1>
  invert maf = 
    let (af1,af2) = fromMAF maf
    in  toMAF (invert af2, invert af1)

  forallInvert maf = 
    let (af1,af2) = fromMAF maf
    in  toMAF (forallInvert af1, forallInvert af2)

-- comp maf1 maf2 = maf2 o maf1
  comp maf1 maf2 = 
    let (af1,ag1) = fromMAF maf1
        (af2,ag2) = fromMAF maf2  
    in  toMAF (comp af1 af2, comp ag1 ag2)

-- Identity MAF
  funcId = toMAF(toAF(1,0),toAF(1,0)) 


-- ============================================
-- Truncated Multi-Valued Affine Functions
-- ============================================
-- a TMAF is a triple (F, S, J) s.t. tamf(I) = F(I /\ S) /\ J

newtype TMAF = TMAF (MAF, Interval RealPlus, Interval RealPlus)
  deriving Eq

instance Show TMAF where
  show (TMAF (maf,s,j)) = 
    show maf++"/\\"++show j++" if x in "++show s

type UnappliedTMAF = Interval RealPlus -> Interval RealPlus -> TMAF

toTMAF :: MAF -> UnappliedTMAF
toTMAF maf s j = TMAF (maf, s, j)

-- Projection: to obtain the MAF from TMAF 
fromTMAF2MAF :: TMAF -> MAF
fromTMAF2MAF (TMAF (maf, s, j)) = maf

-- Projection: to obtain interval J and S from TMAF 
fromTMAF2S, fromTMAF2J :: TMAF -> Interval RealPlus
fromTMAF2S (TMAF (maf, s, j)) = s
fromTMAF2J (TMAF (maf, s, j)) = j

fromTMAF2SJ:: TMAF -> (Interval RealPlus, Interval RealPlus)
fromTMAF2SJ (TMAF (_,s, j)) = (s,j)


instance Applicable TMAF (Interval RealPlus) where
  apply (TMAF (f,s,j)) = 
    \i -> (apply f (i /\ s)) /\ j
  
instance Function TMAF where

-- invert tmaf = (invert maf) /\ S if I \in J 
  invert (TMAF (f,s,j)) = toTMAF (invert f) j s
  forallInvert (TMAF (f,s,j)) = toTMAF (forallInvert f) j s

-- comp tmaf1 tmaf2 = tamf2 o tamf1
  comp (TMAF (f1,s1,j1)) (TMAF (f2,s2,j2)) =
      toTMAF (comp f1 f2) 
             (s1 /\ apply (invert f1) (j1 /\ s2))
             (j2 /\ apply f2 (j1 /\ s2))

-- Identity MAF
  funcId = toTMAF funcId (toInterval (open InfiNeg, open InfiPos)) (toInterval (open InfiNeg, open InfiPos))


-- Fixing flipping intervals
fixInterval i 
  | not (isNormalized i) = point (1.0::Double) - i
  | otherwise            = i


