module ConcretePath (getTrace, concretiseSignature, getArea, interval2extremes) where

import Maybe 
import List

import Input
import Vector
import Infinity
import Interval
import Graph
import Signature

type ConcretePath = [(EdgeId, Interval RealPlus)]


getArea :: Input -> [((EdgeId, Interval RealPlus), (EdgeId, Interval RealPlus))] 
                 -> [((Vector, Vector), (Vector, Vector))]
getArea input areas =
  [ (fromJust $ interval2extremes input e0, fromJust $ interval2extremes input e1) 
  | (e0, e1) <- areas
  , isJust (interval2extremes input e0)
  , isJust (interval2extremes input e1)
  ]

getTrace :: Input -> ConcretePath -> Maybe [(Vector, Vector)]
getTrace input path 
  | Nothing `elem` result = Nothing
  | otherwise             = Just (map fromJust result)
  where
    result = map (interval2extremes input) path 

interval2extremes input (edgeid, interval) 
  | isEmpty interval = Nothing
  | otherwise        = Just (edgeFormula l, edgeFormula u)
  where
    edge = fromJust $ lookup edgeid (inputEdges input)
    (l,u) = values interval
    edgeFormula = 
      \lambda -> position edge + (unitDir edge * constant' lambda)

limitBackwardsPath :: Graph -> ConcretePath -> ConcretePath
limitBackwardsPath graph cpath = zip edges (reverse result)
  where
    (edges, intervals) = unzip cpath
    tmafs = [ fromJust $ graphTMAF graph src dst | src:dst:_ <- tails edges ]
    
    i0:intervals_r = reverse intervals
    tmafs_r = reverse tmafs

    result = scanl (\interval_after (interval_before, tmaf) -> 
              interval_before /\ apply (invert tmaf) interval_after) 
              i0 (zip intervals_r tmafs_r)
    

concretiseSignature :: Graph -> Signature -> Interval RealPlus -> Interval RealPlus -> Maybe ConcretePath
concretiseSignature graph signature interval_0 interval_f  
  | Nothing `elem` tmafs = Nothing
  | isNothing result     = Nothing
  | otherwise            = Just (limitBackwardsPath graph (fromJust result))
  where
    sig = map fromPath (fromSignature signature)
    sig_ss =  [ p | (n,p) <- zip [1..] sig, n `mod` 2 == 0 ]
    tmafs = [ graphTMAF graph src dst | src:dst:_ <- tails (concat sig) ]
         ++ [ graphTMAF graph (last p) (head p) | p <- sig_ss ]
    result = generateConcretePath graph Nothing sig interval_0 interval_f

applySequential graph edges i0 
  | isEmpty (last intervals)   = Nothing
  | otherwise                  = Just path
  where
    transitions = [ (x,y) | x:y:_ <- tails edges ]
    tmafs = [ fromJust $ graphTMAF graph src dst | (src,dst) <- transitions ]
    intervals = scanl (\interval tmaf -> apply tmaf interval) i0 tmafs
    path = zip edges intervals

applyLoop graph s signature i_0 i_f 
  | isEmpty i_0              = Nothing
  | isJust no_iterations     = no_iterations
  | isNothing iteration_path = Nothing
  | isNothing  after_loop    = Nothing
  | otherwise                = Just (fromJust iteration_path ++ fromJust after_loop)
  where
    last_s    = last s
    tmaf_exit = fromJust $ graphTMAF graph last_s (head (concat signature))
    tmaf_loop = fromJust $ graphTMAF graph last_s (head s)

    i_exit = apply tmaf_exit i_0
    i_loop = apply tmaf_loop i_0

    no_iterations = generateConcretePath graph (Just (last s)) signature i_exit i_f 

    iteration_path = applySequential graph s i_loop 
    last_int  = snd $ last $ fromJust iteration_path

    after_loop = applyLoop graph s signature last_int i_f

generateConcretePath graph _ [r] interval_0 interval_f 
  | isEmpty interval_0                    = Nothing
  | isNothing r_applied                   = Nothing
  | isEmpty (interval_f /\ last_interval) = Nothing
  | otherwise                             = r_applied
  where
    r_applied = applySequential graph r interval_0
    last_interval = snd $ last $ fromJust r_applied

generateConcretePath graph last_edge (r:s:signature) interval_0 interval_f
  | isEmpty interval_0     = Nothing
  | isNothing rss_applied  = Nothing
  | isEmpty last_interval  = Nothing
  | isNothing result       = Nothing
  | otherwise              = Just (fromJust rss_applied ++ fromJust result)
  where
    last_edge'         = case r of { [] -> last_edge; _ -> Just (last r) }
    rss                = if Just (last s) == last_edge' then r++s else r++s++s
    rss_applied        = applySequential graph rss interval_0
    last_interval      = snd $ last $ fromJust rss_applied
    
    result = applyLoop graph s signature last_interval interval_f




