-- Input format to Fig 
-- Gordon Pace - Gerardo Schneider - VERIMAG 
-- 06/08/2001
-- A quick hack to enable viewing the user input files.

module Fig ( 
  figSignatureFile, figPathFile, figKernelFile, figIntersectionFile, figMultipleKernelFile 
) where

import Maybe
import List

import Infinity
import Vector
import UserInput
import FileHandling

type FigLine = (Vector,Vector)
type FigPoly = (FigLine, FigLine)
data DrawInstruction = 
    DrawSignature [[Vector]]
  | DrawPath [FigLine]
  | DrawKernel [FigPoly]
  | DrawIntersections [(FigPoly,FigPoly)]
  | DrawMultiple [(Int, Int, [FigPoly])]
                 [(Int, Int, [(FigPoly,FigPoly)])]

instr2Signature (DrawSignature x) = x
instr2Path (DrawPath x) = x
instr2Kernel (DrawKernel x) = x
instr2Intersections (DrawIntersections x) = x

--
colBLACK = 0 :: Int
colGREEN = 2 :: Int
colRED = 4 :: Int
colBLUE = 1 :: Int

factor :: RealPlus
factor = fromInteger 1200 

figSignatureFile :: UserInput -> String -> [[Vector]] -> Bool -> IO ()
figSignatureFile uinput fig sig deeper =
 do { writeFile fig (figUserInputSignature uinput deeper (DrawSignature sig))
    ; putStrLn ("Generated fig file <"++fig++">") 
 } `catch` \err -> fail ("Cannot write to fig file <"++fig++">") 
      
figPathFile :: UserInput -> String -> [(Vector,Vector)] -> Bool -> IO ()
figPathFile uinput fig path deeper =
 do { writeFile fig (figUserInputSignature uinput deeper (DrawPath path))
    ; putStrLn ("Generated fig file <"++fig++">") 
    } `catch` \err -> fail ("Cannot write to fig file <"++fig++">")

figKernelFile :: UserInput -> String -> [((Vector,Vector),(Vector,Vector))] -> IO ()
figKernelFile uinput fig patches =   
  do { writeFile fig (figUserInputSignature uinput True (DrawKernel patches))
     ; putStrLn ("Generated fig file <"++fig++">")
     } `catch` \err -> fail ("Cannot write to fig file <"++fig++">")

figIntersectionFile uinput fig patches =
  do { writeFile fig (figUserInputSignature uinput True (DrawIntersections patches))
     ; putStrLn ("Generated fig file <"++fig++">")
     } `catch` \err -> fail ("Cannot write to fig file <"++fig++">")

figMultipleKernelFile uinput fig patches intersections = 
  do { writeFile fig 
         (figUserInputSignature uinput True 
            (DrawMultiple patches intersections) 
         )
     ; putStrLn ("Generated fig file <"++fig++">")
     } `catch` \err -> fail ("Cannot write to fig file <"++fig++">")
  


initializeFile = 
  unlines [ 
   "#FIG 3.2",
   "Landscape",
   "Center",
   "Inches",
   "A4",
   "100.00",
   "Single",
   "-2",
   scale 1 ++" 2"
  ]

figUserInputSignature :: UserInput -> Bool -> DrawInstruction -> String
figUserInputSignature userinput deeper instruction = 
  initializeFile ++ regionsFig ++ diffInclusionsFig ++ 
  case instruction of 
    DrawSignature sig -> sigFig
    DrawPath path     -> pathFig
    DrawKernel kernel -> kernelFig kernel 
    DrawIntersections kernel -> intersectionsFig kernel
    DrawMultiple kernels intersections -> manyFig kernels intersections
  where
   sig = instr2Signature instruction 
   path = instr2Path instruction

   regionsFig = "# Regions:\n"++
      concat (map (polylineOrPolygon (colBLACK, 70)) edges)
   diffInclusionsFig = "# Differential inclusions:\n"++
      concat (map (uncurry (arrowline colRED 50 vector_arrow_size)) vectors) 
   sigFig = "# Signature:\n"++
      concat (map (arrowpath colBLUE 80 path_arrow_size) sig_path)
   pathFig = "# Path:\n"++
      concat [ filledpolygon (colBLACK, colGREEN,depth) area
             | (area, depth) <- areas_with_depth ]
   kernelFig kernel = "# Kernel:\n"++
      concat (map (filledpolygonnooutline (colGREEN, 100)) (kernel_areas kernel))
   intersectionsFig kernel = "#Kernel:\n"++
      concat (map (filledpolygonnooutline (colGREEN, 100)) (kernel_intersections kernel))
 
   manyFig kernels intersections = 
      "# Kernels:\n"++concat (
      [ concat (map (filledpolygonnooutline (colour, depth)) (kernel_areas kernel))
      | (depth, colour, kernel) <- kernels 
      ] ++
      [ concat (map (filledpolygonnooutline (colour, depth)) (kernel_intersections kernel))
      | (depth, colour, kernel) <- intersections 
      ] 
      )

   regions = userRegions userinput
  
   -- drawing regions
   vertex n = (vertexTable userinput) !! n
   edgesOfRegion region = map vertex (userRegionVertices region)
   edges = map edgesOfRegion regions 

   -- drawing vectors
   centreAndMinDiameterOfRegion :: UserInputRegion -> (Vector, RealPlus)
   centreAndMinDiameterOfRegion region = 
     let vertices = edgesOfRegion region
         vertices' = if head vertices == last vertices 
                        then init vertices else vertices
         centre   = sumVectors vertices' / (fromInteger $ toInteger $ length vertices')
     in  (centre, foldl1 min (map (distance centre) vertices'))
   (centres, diameters) = unzip $ map centreAndMinDiameterOfRegion regions 
   vector_length = foldl1 min diameters / 2
   vector_arrow_size = vector_length / 7

   regionVectors centre region = 
     let left   = userRegionVFirst region
         right  = userRegionVSecond region
         left'  = constant' vector_length * left / constant' (modulus left)
         right' = constant' vector_length * right / constant' (modulus right)
     in  [ (centre, centre+left'), (centre, centre+right') ]

   vectors = concat $ zipWith regionVectors centres regions
   
   -- drawing signatures
   pairs = concat (map (\path -> [ (p1,p2) | p1:p2:_ <- tails path ]) sig_path)
   min_len = foldl1 min (map (uncurry distance) pairs)
   path_arrow_size = min_len / 7

   sig_path = drawPath sig
   drawPath [] = []
   drawPath [r] = [r]
   drawPath [r,s] = [r++s++[head s]]
   drawPath (r1:s:r2:rest) = 
     (r1++s++[head s]): drawPath ((last s:r2):rest)

   -- drawing a concrete path
   areas = 
     [[v1, v1',v2', v2, v1] | (v1,v2):(v1',v2'):_ <- tails path ]
   areas_with_depth = 
     if deeper then zip areas [100..] else reverse (zip (reverse areas) [100..])

   -- drawing a kernel
   kernel_areas kernel = [ [v1, v1',v2', v2, v1] | ((v1,v2), (v1',v2')) <- kernel ]

   -- drawing an intersection of 2 quads kernel
   kernel_intersections intersections = 
     [ intersect_quads [v1, v1',v2', v2, v1] [u1, u1',u2', u2, u1] 
     | (((v1,v2), (v1',v2')), ((u1,u2), (u1',u2'))) <- intersections 
     ]
     where
       -- intersection of two quadrilaterals
      intersect_quads q1 q2 = 
        concat
        [ sortClosestTo e0 $ concat 
           [ intersect_lines (e0, e1) (f0, f1) | (f0:f1:_) <- tails q2 ]
        | (e0:e1:_) <- tails q1
        ]

      -- intersection of two lines
      intersect_lines (p,p') (q,q') 
        | areColinear p q p' && areColinear p q' p' &&
          areColinear q p q' && areColinear q p' q' = colinear_case
        | isJust meeting_point = if (isOnLine o1 && isOnLine o2) then [pnt] else []
        | otherwise            = []
        where
          meeting_point = intersectLines (Line (p,p')) (Line (q,q'))
          (pnt, o1, o2)   = fromJust meeting_point

          colinear_case 
            | (p `onLine` Line (q,q')) && (p' `onLine` Line (q,q')) = [p,p']
            | (q `onLine` Line (p,p')) && (q' `onLine` Line (p,p')) = [q,q']
            | (p `onLine` Line (q,q'))                              = [p,q']
            | (p' `onLine` Line (q,q'))                             = [p',q]
            | otherwise                                             = []
            
      sortClosestTo p ps = sortBy (\q q' -> compare (distance p q) (distance p q')) ps
          
arrowpath col depth size ps =
  concat [ arrowline col depth size src dst | (src:dst:_) <- tails ps ]

arrowline col depth size p1 p2 =
  "2 1 0 1 "++show col++" 7 "++show depth++" 0 -1 0.000 0 0 -1 1 0 2" +++
  "   1 1 1.00 "++ scale (size / 2) ++ " " ++scale size +++ 
  "   "++point p1 ++ point p2 ++"\n"

polylineOrPolygon ps vs
  | last vs == head vs = polygon ps vs
  | otherwise          = polyline ps vs

polyline (col, depth) vs = 
  "2 1 0 1 "++show col++" 7 "++show depth++
  " 0 -1 0.000 0 0 -1 0 0 "++show (length vs) +++
  concat (map point vs) ++ "\n"

polygon (col, depth) vs = 
  "2 3 0 1 "++show col++" 7 "++show depth++
  " 0 -1 0.000 0 0 -1 0 0 "++show (length vs) +++
  concat (map point vs) ++ "\n"

filledpolygonnooutline = fixdraw filledpolygonnooutline'
filledpolygonnooutline' (in_col, depth) [v1,v2] = polyline (in_col, depth) [v1, v2]
filledpolygonnooutline' (in_col, depth) vs =
  "2 3 0 0 0 "++show in_col++" "++show depth++
  " 0 20 0.000 0 0 -1 0 0 "++show (1+length vs) +++
  concat (map point (vs++[head vs])) ++ "\n"

filledpolygon = fixdraw filledpolygon'
filledpolygon' (out_col, in_col, depth) [v1,v2] = polyline (out_col, depth) [v1, v2]
filledpolygon' (out_col, in_col, depth) vs = 
  "2 3 0 1 "++show out_col++" "++show in_col++" "++show depth++
  " 0 20 0.000 0 0 -1 0 0 "++show (1+length vs) +++
  concat (map point (vs++[head vs])) ++ "\n"



fixdraw func ps vs 
  | length vs' < 2 = ""
  | otherwise      = func ps vs'
  where
    vs' = fix vs

fix vs = fix' (last vs) vs
  where
   fix' _ [] = []
   fix' l (p:ps) 
     | l == p    = fix' l ps
     | otherwise = p:fix' p ps

point v = 
  let (x,y) = fromVector v
  in  scale x ++" "++ scale (-y) ++ " "

scale x = show (round $ toDouble $ x*factor)

line1 +++ line2 = line1 ++ "\n" ++ line2
concat' = foldl (+++) []

