-- FileInput Library for the SPDI application
-- Gordon Pace - VERIMAG 
-- 26/07/2001

module FileHandling (
  fileToUserInput, fileToInput
) where

import Maybe
import Char
import List
import Parser
import Useful
import Infinity
import Input
import Vector
import UserInput 

fileToInput :: FilePath -> 
     IO (Input, String -> Maybe Vector, (String, String) -> [EdgeId], EdgeId -> (String, String))
fileToInput filename = 
  do { userinput <- fileToUserInput filename
     ; let (input, points2edgeID, edgeID2points) = userInputToInput userinput
     ; let point2vector str = lookup str (zip (pointNames userinput) (vertexTable userinput))
     ; return (input, point2vector, points2edgeID, edgeID2points)
     }

fileToUserInput :: FilePath -> IO UserInput
fileToUserInput filename = 
  do { filecontents <- (readFile filename `catch`  \err -> fail ("Cannot read spdi file <"++filename++">"))
     ; let filelines = map removeComments (lines filecontents)
     ; let contents  = filter (hasContent . snd) (zip [1..] filelines)
     ; (pointlines, vectorlines, regionlines) <- splitUp contents
     ; points <- inputPoints "Points" pointlines
     ; vectors <- inputPoints "Vectors" vectorlines
     ; regions <- readRegions points vectors regionlines
     ; convertToInput points regions
     }

removeComments = dropWhile isSpace . fst . break ('*' ==)

convertToInput :: 
  [(String, Vector)] ->
  [([String], (Vector, Vector))] -> IO UserInput
convertToInput points regions 
  | duplicatePointDefinitions /= [] =
      fail ("Duplicate point definition ("++show (head duplicatePointDefinitions)++")")
  | idempotentEdges /= [] =
      fail ("Idempotent edges are not allowed ("++head idempotentEdges++")")
  | otherwise = 
      return $ 
      UserInput {
        pointNames  = map fst points,
        vertexTable = vertexInformation,
        userRegions = map buildRegion regions
      } 
  where
    (vertexLabels, vertexInformation) = unzip points
    (verticesRs, _) = unzip regions

    duplicatePointDefinitions = map snd $ 
      filter fst (zip [p `elem` ps | (p:ps) <- tails vertexInformation] vertexInformation)
    idempotentEdges = concat (map (\vs -> [v | v:w:_ <- tails vs, v==w]) verticesRs)

    buildRegion (vertices, (vfirst, vsecond)) = 
      UserInputRegion {
        userRegionVFirst = vfirst, userRegionVSecond = vsecond,
        userRegionVertices = map index vertices
      }
 
    index vertex_name = fromJust (findIndex (vertex_name ==) vertexLabels)

-- Convert lines to region information
readRegions :: [(String, Vector)] -> [(String, Vector)] -> [(Int, String)] ->
  IO [([String], (Vector, Vector))]
readRegions points vectors [] = return []
readRegions points vectors (region:regions) = 
  do { region'  <- readRegion points vectors region
     ; regions' <- readRegions points vectors regions
     ; return (region':regions')
     }

readRegion :: [(String,Vector)] -> [(String, Vector)] -> (Int, String) ->
  IO ([String], (Vector, Vector))
readRegion points vectors (n, line)
  | rest1 == ""  =
      fail ("Line "++show n++": Region definition line not in the form p1:p2:...:pn, v1, v2")
  | vector1 == "" || vector2 == "" =
      fail ("Line "++show n++": Region definition lines not in the form p1:p2:...:pn, v1, v2")
  | isNothing actual_vector1  =
      fail ("Line "++show n++": Vector labelled \""++vector1'++"\" not defined")
  | isNothing actual_vector2  =
      fail ("Line "++show n++": Vector labelled \""++vector2'++"\" not defined")
  | not (null unknown_labels) =
      fail ("Line "++show n++": Point labelled \""++head unknown_labels++"\" not defined")
  | length polygon_labels < 3 =
      fail ("Line "++show n++": Regions must have at least three vertices")
  | all_collinear             =
      fail ("Line "++show n++": All points in region are collinear")
  | not (null nongoodEdges)   =
      fail ("Line "++show n++": Region fails goodness property (edge "++head nongoodEdges++" lies within vectors)")
  | otherwise = 
      do { if inverted_vectors 
             then putStrLn ("Warning! Line "++show n++": Exchanging vectors to be in the correct order.")
             else putStr ""
         ; return (polygon_labels, (actual_vector1', actual_vector2'))
         }
  where
    (polygon, rest1) = break (',' ==) line
    polygon' = clean polygon

    (vector1, vector2) = break (',' ==) (tail rest1)
    vector1' = clean vector1
    vector2' = clean (tail vector2)
     
    actual_vector1 = lookup vector1' vectors
    actual_vector2 = lookup vector2' vectors

    inverted_vectors = (fromJust actual_vector2) `isClockwiseOf` (fromJust actual_vector1) 
    (actual_vector1', actual_vector2') =
       if inverted_vectors  
         then (fromJust actual_vector2, fromJust actual_vector1)
         else (fromJust actual_vector1, fromJust actual_vector2)

    polygon_labels = map clean $ lines (map (\x -> if x `elem` ":!?" then '\n' else x) polygon')

    polygon_vertices = map (`lookup` points) polygon_labels
    unknown_labels = [ vname | (vname, Nothing) <- zip polygon_labels polygon_vertices ]
    all_collinear = and 
      [ areColinear v1 v2 v3 | v1:v2:v3:_ <- tails (map fromJust polygon_vertices) ]

    polygon_vertices' = map fromJust polygon_vertices
    edge_names = [e++"-"++e' | e:e':_ <- tails polygon_labels]
    edge_vectors = [v'-v | v:v':_ <- tails polygon_vertices']
    isBad vector = ((vector `isClockwiseOfOrOn` actual_vector2') && 
                     (vector `isAnticlockwiseOfOrOn` actual_vector1'))||
                    (((-vector) `isClockwiseOfOrOn` actual_vector2') &&
                     ((-vector) `isAnticlockwiseOfOrOn` actual_vector1'))

    nongoodEdges = [name | (name,v) <- zip edge_names edge_vectors, isBad v]
    

-- Read points from a list of lines
inputPoints :: String -> [(Int, String)] -> IO [(String, Vector)]
inputPoints section lines = inputPointsAux [] section lines 
  where
    inputPointsAux :: [String] -> String -> [(Int, String)] -> IO [(String, Vector)]
    inputPointsAux defined section [] = return []
    inputPointsAux defined section (line:lines) =
      do { (label, position) <- inputPoint defined section line
         ; rest <- inputPointsAux (label:defined) section lines
         ; return ((label, position):rest)
         }

inputPoint :: [String] -> String -> (Int, String) -> IO (String, Vector)
inputPoint defined section (n, line)
  | rest1 == ""  = 
      fail ("Line "++show n++": "++section ++ " lines must start with a label followed by a dot")
  | label `elem` defined =
      fail ("Line "++show n++": Non-unique label "++ label ++" in section "++section)
  | number2 == "" = 
      fail ("Line "++show n++": "++section ++ "("++ label' ++")" ++
             "Coordinates must be a pair of numbers separated by a comma")
  | isNothing number1' =
      fail ("Line "++show n++": Invalid x-coordinate in point definition of <"++label'++"> in " ++ section)
  | isNothing number2' =
      fail ("Line "++show n++": Invalid y-coordinate in point definition of <"++label'++"> in " ++ section)
  | otherwise = 
      return (label', Vector (fromJust number1', fromJust number2'))
  where
    (label, rest1) = break ('.' ==) line
    label' = clean label
    (number1, number2) = break (',' ==) (tail rest1)
    (number1', number2') = (isNumber number1, isNumber (tail number2))

-- remove front and trailing whitespace
clean :: String -> String
clean s = reverse (dropWhile isSpace (reverse (dropWhile isSpace s)))
  
-- Is this a content line (comments removed earlier)?
hasContent xs = any (not . isSpace) xs

-- Do two strings match (the second may contain trailing white space)
s1 `isString` s2 = dropWhile isSpace (reverse s1) == dropWhile isSpace (reverse s2)

-- Split up lines into three sets: points, vectors, regions
splitUp lines = 
  let (_,others) = break (isString "Points:" . snd) lines
  in  splitUp1 others

splitUp1 [] = fail "File does not start with \"Points:\" directive"
splitUp1 (_:rest) = 
  do { (vectorlines, regionlines) <- splitUp2 others; return (pointlines, vectorlines, regionlines) } 
  where 
    (pointlines, others) = break (isString "Vectors:" . snd) rest

splitUp2 [] = fail "File does not have a \"Vectors:\" directive"
splitUp2 (_:rest) =
  do { regionlines <- splitUp3 others; return (vectorlines, regionlines) }
  where
    (vectorlines, others) = break (isString "Regions:" . snd) rest

splitUp3 [] = fail "File does not have a \"Regions:\" directive"
splitUp3 (_:rest) = return rest


