module Generator (
  generateSignatures, graphToRelation, allSimpleLoops
) where

import Maybe
import List
import Array

import Signature
import Interval
import Graph
import Successor

-- Generate the signature types from a graph

-- 1. Produce the transition relation 
type TransitionRelation = [(EdgeId,EdgeId)]
realline = toInterval (closed InfiNeg, closed InfiPos) :: Interval RealPlus

graphToRelation :: Graph -> TransitionRelation
graphToRelation graph = 
  -- DEBUG: Show the transitions before optimization
  -- error $ show [ (from,to) | from <- edges, to <- edges, isJust $ graphTMAF graph from to]
  -- DEBUG: Show the resultant transitions after optimization (one step application)
  -- error $ show [ (from,to) | from <- edges, to <- edges, possible from to ]
  [ (from,to) | from <- edges, to <- edges, possible from to ] 
  where
    edges = graphDomain graph

    possible from to = 
      let tmaf = graphTMAF graph from to
      in  isJust tmaf && not (isEmpty (apply (fromJust tmaf) realline))


graphToRelationOptimized :: Graph -> (EdgeId, EdgeId) -> TransitionRelation
graphToRelationOptimized graph (src, dst) = 
  -- DEBUG: Show unoptimised relations
  -- error $ show $ length relation0
  -- DEBUG: Show optimised relations
  -- error $ show $ relation' relation0
  -- error $ show $ map length $ relation'' relation0
  relation' relation0
  where
    edges = graphDomain graph

    -- remove transitions which are Nothing or Just tmaf, which when applied gives emptyInterval.
    possible from to = 
      let tmaf = graphTMAF graph from to
      in  isJust tmaf && not (isEmpty (apply (fromJust tmaf) realline)) 
  
    relation0 = [ (from,to) | from <- edges, to <- edges, possible from to ] 

    -- fixed point to discard sinks and top positions not otherwise reachable
    relation' rel
      | length rel == length rel' = rel
      | otherwise                 = relation' rel'
      where
        rel' = [(from, to) | (from,to) <- rel, from `notElem` tops rel && to `notElem` sinks rel ]
    sinks rel = [ e | e <- edges ,e `notElem` map fst rel, e /= dst ]
    tops  rel = [ e | e <- edges, e `notElem` map snd rel, e /= src ]
       
    relation'' rel = rel:relation'' rel'
      where
        rel' = [(from, to) | (from,to) <- rel, from `notElem` tops rel && to `notElem` sinks rel ]

-- 2. Generate the loops finishing on a certain edge
loopsFinish :: Graph -> TransitionRelation -> EdgeId -> [Path]
loopsFinish graph transitions finish = map toPath (loopsFinishAux [finish] funcId)
  where
    loopsFinishAux edges@(e:_) tmaf 
      | isEmpty (apply tmaf realline) = []
      | otherwise                     =
          (if (finish, e) `elem` transitions then [edges] else []) ++
          concat [ loopsFinishAux (p:edges) (comp (fromJust $ graphTMAF graph p e) tmaf) 
                 | (p,e') <- transitions, e'==e, p `notElem` edges 
                 ]

-- 2b. Generate all loops

allSimpleLoops :: Graph -> [Path]
allSimpleLoops graph = 
  nubBy sameButShifted $ concat [ loopsFinish graph transitions finish | finish <- graphDomain graph]
  where
    tmaf e e' = fromJust $ graphTMAF graph e e'

    transitions = graphToRelation graph 


-- 3. Generate a list of all loops with their exiting edge, and a list of the loops starting and
--    exiting on a particular edge (the destination)
loops graph transitions domain dst = (finish_loops, loop_lookup)
  where
    max_element = maximum domain

    lookupRotation x = listToMaybe . map fst  . filter (sameButShifted x . snd) 
    labelTable' :: Int -> [(Int,Path)] -> [(EdgeId, [Path])] -> [(EdgeId, [(Int, Path, [Interval (Extend Double)])])]
    labelTable' n _ [] = []
    labelTable' n known ((exit, loops):rest) = 
      let ((n', known'), loops') =
             mapAccumL (\(n,known) loop -> 
                          case lookupRotation loop known of
                            Nothing      ->  ((n+1, (n,loop):known), info n       loop)
                            Just n_known ->  ((n  ,          known), info n_known loop)
                       ) (n,known) loops
          info n loop = (n, loop, filter (not . isEmpty) $ exitLoop1 graph loop exit realline)
      in  (exit, loops'): labelTable' n' known' rest

    loop_array0 = array (0,max_element) [ (edge, loopsFinish graph transitions edge) | edge <- [0..max_element] ]
    loop_table = labelTable' 0 []
             [ (   edge
               ,   filter (feasibleLoop graph edge) 
                 $ filter (([] ==) . ([edge,dst] `intersect`) . fromPath) 
                 $ concat
                 [ loop_array0 ! pre
                 | (pre, edge') <- transitions, edge'==edge
                 ]
               )
             | edge <- [0..max_element]
             ]
 
    loop_array = array (0,max_element) loop_table
    loop_lookup edge = loop_array ! edge
    finish_loops = filter (feasibleLoop graph dst) $ map lastToHead $ loopsFinish graph transitions dst

    lastToHead xs = lastPath xs !: initPath xs

-- 4. Generate the signatures

-- 5. Generate the signatures more efficiently
generateSignatures :: Graph -> EdgeId -> EdgeId -> ([(EdgeId,EdgeId)],[Signature],[Path])
generateSignatures graph src dst = 
  (transitions, map toSignature signatures, finish_loops)
  where
    transitions = graphToRelationOptimized graph (src,dst)
    (finish_loops, loop_lookup) = loops graph transitions (graphDomain graph) dst

    signatures = generateSignaturesAux ([toPath [dst]], [dst], [], funcId)

    generateSignaturesAux (future, rs, ss, tmaf)
      | first == src = [future]
      | otherwise    = concat (map generateSignaturesAux results)
      where
        first = futureHead future
        r1:s1:r2:rest = future++[toPath[],toPath[]]

        all_but_not_first_s_and_following_r  = concat (map fromPath (r1:rest))
        all_but_not_first_rs = concat (map fromPath rest)
    
        extend_r = [ ( consEdge pre future -- extend the signature along an r 
                     , pre:rs, ss
                     , comp (fromJust $ graphTMAF graph pre first) tmaf
                     )
                   | (pre, edge) <- transitions
                   , edge == first              -- extend one step back to x such that
                   , pre `notElem` rs           -- x not elem of any r (P3)
                   -- , pre `notElem` all_but_not_first_s_and_following_r
                                 -- x not elem of any r or s (except the next s and the following r)
                   , (pre !: r1) /= s1          -- the first loop is not repeated (P7)
                   ]

        extend_r' = 
          filter (\(path, _, _, tmaf) -> 
                      (not $ isEmpty $ apply tmaf realline) && 
                      (canWork (toSignature path) realline)) extend_r 
          -- remove seq extensions where applying the path leads to a dead-end
    
        extend_s = [ ( toPath []:loop:future  -- extend by adding a loop
                     , rs
                     , label:ss
                     , funcId
                     )
                   | (label, loop, intervals) <- loop_lookup first
                          -- choose a loop exiting on first
                   --, all_but_not_first_rs `intersect` fromPath loop == [] 
                          -- no common edges with future 
                   , not (label `elem` ss) -- which has not yet been used (P4) and 
                   , fromPath loop `intersect` fromPath r1 == [] || fromPath s1 == []
                          -- has no common element with the first r segment (P5)
                          -- the second option is to cover the fact that this property
                          -- does not hold on the last s,r pair.
                   , or (map (canWork sig) intervals) -- filter 
                   ]
         
        sig = toSignature future
        canWork sig interval  = not $ areEmpty $ applySignature graph interval sig

        results = extend_r' ++ extend_s

 
-- extract the r and s parts of the signatures
signatureRSs :: [Path] -> ([Path], [Path])
signatureRSs [r] = ([r],[toPath []])
signatureRSs (r:s:sig) = let (rs,ss) = signatureRSs sig in (r:rs,s:ss)
signatureRSs _ = error "Internal error: signatureRSs"

-- the first edge in the future
futureHead :: [Path] -> EdgeId
futureHead (Path []:Path (x:_):_) = x
futureHead (Path (x:_):_)    = x
futureHead paths = error ("Internal error: futureHead -> "++show paths)

-- cons-ing an edge to the future
consEdge :: EdgeId -> [Path] -> [Path]
consEdge x (xs:xss) = (x !: xs):xss

-- are two paths equivalent but possible rotated? We assume that the paths contain no duplicates
sameButShifted :: Path -> Path -> Bool
p1 `sameButShifted` p2 =
  let p1_head = headPath p1
      (p2before, p2after) = break (p1_head ==) (fromPath p2)
  in  p2after++p2before == fromPath p1


