-- Infinity Library
-- Gordon Pace - Gerardo Schneider - VERIMAG 
-- 16/07/2001

module Infinity (
  Extend(..), 
  RealPlus, sqrtPlus, toDouble
) where

import Numeric
import Useful

-- an Extend a can be Infinite (positive or negative) or finite
data Extend a =
    InfiPos
  | InfiNeg
  | Finite a
  deriving Eq

fromExtend :: Extend a -> a
fromExtend (Finite x) = x
fromExtend InfiPos = error "fromExtend applied +infinity"
fromExtend InfiNeg = error "fromExtend applied -infinity"

-- If a is an ordered type, then we define order for Extend a
instance (Ord a) => Ord (Extend a) where
  {-# SPECIALIZE instance Ord (Extend Double) #-}
  _ <= InfiPos = True
  InfiNeg <= _ = True
  (Finite x) <= (Finite y) = x <= y
  _ <= _        = False

instance Show a => Show (Extend a) where
  {-# SPECIALIZE instance Show (Extend Double) #-}
  show InfiPos = "+infinity" 
  show InfiNeg = "-infinity"
  show (Finite x) = show x

instance Read a => Read (Extend a) where
  {-# SPECIALIZE instance Read (Extend Double) #-}
  readsPrec n string = 
    map (\(n,rest) -> (Finite n, rest)) (readsPrec n string)

-- Operation over Extend a whenever a is a numerical type
instance Num a => Num (Extend a) where
  {-# SPECIALIZE instance Num (Extend Double) #-}
  negate InfiPos = InfiNeg
  negate InfiNeg = InfiPos 
  negate (Finite f) = Finite (negate f) 

  InfiPos + InfiNeg = error "(infinity - infinity) undefined"
  InfiNeg + InfiPos = error "(infinity - infinity) undefined"
  InfiPos + _       = InfiPos
  _ + InfiPos       = InfiPos
  InfiNeg + _       = InfiNeg
  _ + InfiNeg       = InfiNeg
  (Finite x) + (Finite y) = Finite (x+y)

  Finite x * Finite y = Finite (x*y)
  x * y = if signum x * signum y == Finite 1 then InfiPos else InfiNeg

  signum InfiPos    = Finite 1
  signum InfiNeg    = Finite (-1)
  signum (Finite x) = Finite (signum x)
  
  abs (Finite x) = Finite (abs x)
  abs _          = InfiPos

  fromInteger  =  Finite . fromInteger

-- Operation over Extend a whenever a is a Fractional type
instance Fractional a => Fractional (Extend a) where
  {-# SPECIALIZE instance Fractional (Extend Double) #-}
  InfiPos / InfiNeg = error "Dividing infinite value by another infinite value"
  InfiNeg / InfiPos = error "Dividing infinite value by another infinite value"
  InfiPos / x       = signum x * InfiPos
  InfiNeg / x       = signum x * InfiNeg
  Finite x / Finite y 
    | y == 0    = signum (Finite x) * InfiPos
    | otherwise = Finite (x / y)
  _ / _ = Finite 0

  fromRational = Finite . fromRational

-- Extended Real Numbers
type RealPlus = Extend Double

-- Square root
sqrtPlus :: RealPlus -> RealPlus
sqrtPlus (Finite x) = Finite (sqrt x)
sqrtPlus InfiPos    = InfiPos
sqrtPlus _          = error "Square root of InfiNeg not computable"

-- Get the actual finite value 
toDouble :: RealPlus -> Double
toDouble = fromExtend


