-- Input Library for the SPDI application
-- Gordon Pace - VERIMAG 
-- 26/07/2001

module Input (
  module Vector, 
  module AffineFunctions,

  Parity(..), opposite, 

  Input(..), position, unitDir, lambdaRange,
  Region(..), vSecond, vFirst, edgeInfo, 
  EdgeId, Edge(..), inputEdges, inputRegions,

  getEdge,
  regionOf, countEdges,
  edgeToEdgeTMAF, edgeToEdgeMAF
) where

import Maybe
import Infinity
import Interval
import Vector
import AffineFunctions

parityIn = In
parityOut = Out

-- an edge can be of Input (In) or Output (Out) relative to a region
data Parity = In | Out deriving (Eq, Enum, Show)

opposite In  = Out
opposite Out = In

-- Edge identifier
type EdgeId     = Int

-- An edge is characterised by one extremity (position), a director 
-- vector (unitDir) and the maximum lambda:
-- { position + alpha.uniDir | alpha \in lambdaRange }
data Edge = 
  Edge {
    position, unitDir :: Vector,
    lambdaRange :: Interval RealPlus
  } deriving Eq

instance Show Edge where
  show e = "{ "++show(position e)++"+lambda." ++show(unitDir e)++
           " | lambda <- "++show (lambdaRange e)++" }"

-- A region is determined by the differential inclusion (two vectors 
-- vSecond and vFirst) and a list of edges (identifier and Parity)
data Region  = 
  Region {
     vSecond, vFirst :: Vector,
     edgeInfo :: [(EdgeId, Parity)]
  } deriving (Eq, Show)


-- This is the internal Input data: a list of Edge ids and internal 
-- edge representation (Edge) with a list of Regions.
data Input = 
  Input { 
    inputEdges :: [(EdgeId, Edge)],
    inputRegions :: [Region]
  } deriving (Eq, Show)

-- Given an Input and an Edge id, it returns the Edge information
getEdge :: Input -> EdgeId -> Edge
getEdge input edge_id = 
  fromJust (lookup edge_id (inputEdges input))

-- Given two edges it returns the corresponding unapplied TMAF or Nothing.
edgeToEdgeTMAF :: 
  Input -> EdgeId -> EdgeId -> Maybe TMAF
edgeToEdgeTMAF input e1 e2 = 
  case maybe_maf of
    Nothing  -> Nothing
    _        -> Just tmaf
  where
    maybe_maf = edgeToEdgeMAF input e1 e2
    maf = fromJust maybe_maf

    interval1 = lambdaRange (fromJust (lookup e1 (inputEdges input)))
    interval2 = lambdaRange (fromJust (lookup e2 (inputEdges input)))

    s = apply (invert maf) interval2 /\ interval1
    j = apply maf interval1 /\ interval2

    tmaf = toTMAF maf s j

-- It computes the MAF of two edges (using the information of the 
-- Input: region and vectors (diff. inclusions)): i.e. the one-step 
-- successor
edgeToEdgeMAF :: Input -> EdgeId -> EdgeId -> Maybe MAF
edgeToEdgeMAF input src_id dst_id = 
  case maybeRegion of
    Nothing -> Nothing
    _       -> Just (toMAF (af1, af2))
  where
    maybeRegion = regionOf input src_id dst_id

    region = fromJust maybeRegion
 
    (v_second, v_first) = (vSecond region, vFirst region)
    
    src_edge = getEdge input src_id
    dst_edge = getEdge input dst_id

    e1 = unitDir src_edge
    e2 = unitDir dst_edge

    hat_v_second  = hat v_second
    hat_v_first = hat v_first

    v  = position src_edge
    v' = position dst_edge

    v_v' = v - v'
    e2_dot_hat_v_second  = e2 >.< hat_v_second
    e2_dot_hat_v_first = e2 >.< hat_v_first

    af1 = toAF ((e1 >.< hat_v_second) / e2_dot_hat_v_second, 
                (v_v' >.< hat_v_second) / e2_dot_hat_v_second)
    af2 = toAF ((e1 >.< hat_v_first) / e2_dot_hat_v_first,    
                (v_v' >.< hat_v_first) / e2_dot_hat_v_first)

-- The number of distinct edges. We assume that edge ids are 
-- allocated sequentially, starting from 0, and hence, the maximum
-- id used is one less the edge count. 
countEdges :: Input -> Int
countEdges input = 1+maximum (-1:map fst (inputEdges input))

-- The (unique) region in which we can go from edge src to
-- edge dst. 
regionOf :: Input -> EdgeId -> EdgeId -> Maybe Region
regionOf input src dst =
  choose (inputRegions input)
  where
    choose [] = Nothing
    choose (r:rs)
      | (src,In)  `elem` es &&
        (dst,Out) `elem` es    = Just r
      | otherwise              = choose rs
      where
        es = edgeInfo r

