-- Interval Library
-- Gordon Pace - Gerardo Schneider - VERIMAG 
-- 16/07/2001

-- This version is for generic types Intervals

module Interval (
  Extremity, value, values, isOpen, isClosed, clop, closed, open,
  Interval, fromInterval, toInterval, toClosedInterval, toOpenInterval,
  emptyInt, isEmpty, areEmpty,
  normalize, isNormalized, isIn, subsetOf, subsetEq,
  priorityClOp, (/\), point, overlaps
) where

import Infinity

-- Extremity gives the extremes of an interval with the corresponding
-- bracket (Open or Closed)
data Extremity a = 
    Open a
  | Closed a 
  deriving (Eq, Show)

-- An interval can be EmptyInterval or Interval (l,u)
data Interval a =
    EmptyInterval
  | Interval (Extremity a, Extremity a)
  deriving (Eq)

instance (Show a) => Show (Interval a) where
  {-# SPECIALIZE instance Show (Interval (Extend Double)) #-}
  show EmptyInterval = "Empty"
  show (Interval (Open x, Closed y)) = "(" ++ show x ++ "," ++ show y ++ "]"
  show (Interval (Closed x, Open y)) = "[" ++ show x ++ "," ++ show y ++ ")"
  show (Interval (Closed x, Closed y)) = "[" ++ show x ++ "," ++ show y ++ "]"
  show (Interval (Open x, Open y)) = "(" ++ show x ++ "," ++ show y ++ ")"
  

-- Constructor & Projections

fromInterval :: Interval a -> (Extremity a,Extremity a)
fromInterval (Interval (l,u)) = (l,u)

value :: Extremity a -> a
value (Open x) = x
value (Closed x) = x

values :: Interval a -> (a,a)
values interval = let (l,u) = fromInterval interval in (value l, value u)
 
clop :: Extremity a -> b -> Extremity b
clop (Open x) = Open 
clop (Closed x) = Closed 

open, closed :: a -> Extremity a
open = Open
closed = Closed

isOpen :: Extremity a -> Bool
isOpen (Open _) = True
isOpen (Closed _) = False

isClosed :: Extremity a -> Bool
isClosed (Closed _) = True
isClosed (Open _) = False

toInterval :: (Extremity a,Extremity a) -> (Interval a)
toInterval (l,u) = Interval (l,u)

toClosedInterval, toOpenInterval :: (a, a) -> (Interval a)
toClosedInterval (l,u) = Interval (Closed l,Closed u)
toOpenInterval (l,u) = Interval (Open l,Open u)


emptyInt :: Interval a
emptyInt = EmptyInterval

isEmpty :: Ord a => Interval a -> Bool
isEmpty  = (emptyInt ==) . normalize
-- isEmpty i = (normalize i == emptyInt)

areEmpty :: Ord a => [Interval a] -> Bool
areEmpty = all isEmpty

-- Normalization
-- normalize (3,3) = Empty
-- normalize [5,3] = Empty

isNormalized :: Ord a => Interval a -> Bool
isNormalized (Interval (l,u)) = 
  u > l || (u == l && isClosed u && isClosed l)

normalize :: Ord a => Interval a -> Interval a
-- normalize EmptyInterval = emptyInt
normalize (Interval (l,u))
  | value l < value u                              = Interval (l,u) -- efficiency
  | value l > value u                              = emptyInt
  | (value l == value u) && (isOpen l || isOpen u) = emptyInt
  | otherwise                                      = Interval (l,u)
normalize EmptyInterval = EmptyInterval

-- Operations

-- Open extremities have priority over Closed ones
priorityClOp :: Extremity a -> Extremity a -> a -> Extremity a
priorityClOp (Closed _) (Closed _) = Closed
priorityClOp _ _                   = Open


-- If a is an ordered type then we define min and max for Extremity a
instance Ord a => Ord (Extremity a) where 
  {-# SPECIALIZE instance Ord (Extremity (Extend Double)) #-}
  compare x y = compare (value x) (value y)

  min x y 
    | value x < value y = x
    | value x > value y = y
    | otherwise = priorityClOp x y (value x)

  max x y 
    | value x < value y = y
    | value x > value y = x
    | otherwise = priorityClOp x y (value x)       

-- We can also perform arithmetic operations on extremities
instance Num a => Num (Extremity a) where
  {-# SPECIALIZE instance Num (Extremity (Extend Double)) #-}
  x + y = priorityClOp x y (value x + value y)
  x - y = priorityClOp x y (value x - value y)
  x * y = priorityClOp x y (value x * value y)
  negate x = clop x (negate (value x))

  signum x = clop x (signum (value x))
  abs x    = clop x (abs    (value x))

  fromInteger x = Closed (fromInteger x)

-- We can also perform arithmetic operations on intervals
instance Num a => Num (Interval a) where
  {-# SPECIALIZE instance Num (Interval (Extend Double)) #-}
  x + y = toInterval (lx+ly, rx+ry)
    where
     (lx,rx) = fromInterval x
     (ly,ry) = fromInterval y
  x - y = toInterval (lx-ly, rx-ry)
    where
     (lx,rx) = fromInterval x
     (ly,ry) = fromInterval y
  x * y = toInterval (lx*ly, rx*ry)
    where
     (lx,rx) = fromInterval x
     (ly,ry) = fromInterval y
  negate x = toInterval (negate l, negate r)
    where 
      (l,r) = fromInterval x

  abs x = toInterval (abs l, abs r)
    where
      (l,r) = fromInterval x   
  signum x = toInterval (signum l, signum r)
    where
      (l,r) = fromInterval x
  fromInteger x = toInterval (fromInteger x,fromInteger x)
 

-- intersection
(/\) :: Ord a => Interval a -> Interval a -> Interval a
i /\ j =
  case (i,j) of
    (Interval (li,ui), Interval (lj, uj)) -> Interval (max li lj, min ui uj)
    _                                     -> emptyInt
        
-- a point interval
point :: a -> Interval a
point x = toInterval (Closed x, Closed x)

-- Check whether a point is in the interval
isIn :: Ord a => a -> Interval a -> Bool
isIn x i = not (isEmpty (point_x /\ i))
  where
    point_x = point x

-- Subset i j iff i is a subset of j (for normalized intervals)
subsetOf :: Ord a => Interval a -> Interval a -> Bool
subsetOf i j = (subsetEq i_norm j_norm) && not (i_norm == j_norm)
  where 
    i_norm = normalize i
    j_norm = normalize j

-- Subset or equal (for normalized intervals)
subsetEq :: Ord a => Interval a -> Interval a -> Bool
subsetEq i j = (i_norm /\ j_norm) == i_norm
  where 
    i_norm = normalize i
    j_norm = normalize j

-- Overlapping intervals
overlaps :: Ord a => Interval a -> Interval a -> Bool
i1 `overlaps` i2 = not(isEmpty (i1 /\ i2))

    
