module Kernel (invariance, viability, controllability) where 

import List
import Maybe

import Interval
import Graph
import Signature
import Successor

type Polygon = ((EdgeId, Interval RealPlus), (EdgeId, Interval RealPlus))

loopTMAF graph p@(Path (es@(e0:_))) =
    mcomp [ tmafLookup graph e e' | e:e':_ <- tails (es++[e0]) ]

invariance :: Graph -> Path -> [Polygon]
invariance graph p = 
  concat $ preTildeBar graph p interval
  where
    loop = loopTMAF graph p

    interval =  apply (forallInvert loop) (fromTMAF2J loop)
    
viability :: Graph -> Path -> [Polygon]
viability graph p = 
  --error $ show $ loopTMAF graph p
  concat $ preBar graph p s
  where
    s = fromTMAF2S $ loopTMAF graph p

controllability :: Graph -> Path -> [(Polygon, Polygon)]
controllability graph p =
  concat 
    [ [ (p,s) | p <- ps, s <- ss ]
    | (ps, ss) <- zip pres succs 
    ]
  where
    pres  = preBar  graph p interval
    succs = succBar graph p interval

    (l_loop,u_loop) = values $ uncurry (/\) $ fromTMAF2SJ $ loopTMAF graph p

    interval = toClosedInterval $ 
      case fromJust $ loopType graph p of
        Stay (l,u)  -> (l,u)
        ExitLeft u  -> (l_loop, u)
        ExitRight l -> (l, u_loop)
        ExitBoth    -> (l_loop, u_loop)


preBar :: Graph -> Path -> Interval RealPlus -> [[Polygon]]
preBar graph (Path (es@(e0:_))) interval = 
  reverse $ preBarAux interval (e0: reverse es)
  where
    tmaf = tmafLookup graph

    preBarAux interval (e0:e1:es) =
      let pre_interval = apply (invert (tmaf e1 e0)) interval 
          full_pre     = 
             [ ((e, apply (invert (tmaf e e0)) interval), (e0, interval))
             | e <- graphDomain graph
             , isJust (graphTMAF graph e e0) 
             , not $ isEmpty $ apply (invert (tmaf e e0)) interval
             ]
      in  full_pre: preBarAux pre_interval (e1:es)

    preBarAux _ _ = []

preTildeBar :: Graph -> Path -> Interval RealPlus -> [[Polygon]]
preTildeBar graph (Path (es@(e0:_))) interval = 
  reverse $ preTildeBarAux interval (e0:reverse es)
  where
    tmaf = tmafLookup graph

    preTildeBarAux interval (e0:e1:es) =
      let pre_interval = apply (forallInvert (tmaf e1 e0)) interval 
          full_pre     = 
             [ ((e, apply (forallInvert (tmaf e e0)) interval), (e0, interval))
             | e <- graphDomain graph
             , isJust (graphTMAF graph e e0) 
             , not $ isEmpty $ apply (forallInvert (tmaf e e0)) interval
             ]
      in  full_pre: preTildeBarAux pre_interval (e1:es)

    preTildeBarAux _ _ = []

succBar :: Graph -> Path -> Interval RealPlus -> [[Polygon]]
succBar graph (Path (es@(e0:_))) interval =
  succBarAux interval (es++[e0])
  where
    tmaf = tmafLookup graph

    succBarAux interval (e0:e1:es) = 
      let succ_interval = apply (tmaf e0 e1) interval
          full_succ = 
            [ ((e0, interval), (e, apply (tmaf e0 e) interval)) 
            | e <- graphDomain graph
            , isJust (graphTMAF graph e0 e)
            , not $ isEmpty $ apply (tmaf e0 e) interval
            ]
      in full_succ: succBarAux succ_interval (e1:es)

    succBarAux _ _ = []

tmafLookup graph e e' = fromJust $ graphTMAF graph e e'
