-- Input format to Postscript code
-- Gordon Pace - Gerardo Schneider - VERIMAG 
-- 06/08/2001
-- A quick hack to enable viewing the user input files.

module PS ( 
  psSPDIFile, psEdgesFile
) where

import Maybe
import List
import Infinity
import Vector
import UserInput
import FileHandling

xsize, ysize, margin :: RealPlus
xsize =  8*72 - 2 * margin
ysize = 11*72 - 2 * margin
margin = 72

psEdgesFile :: UserInput -> String -> String -> [[(Vector)]] -> IO ()
psEdgesFile uinput spdi ps sig =
 do { writeFile ps (psUserInputSignature spdi uinput sig);
      putStrLn ("Generated ps file <"++ps++">") 
 } `catch` \err -> fail ("Cannot write to ps file <"++ps++">") 
      

psSPDIFile :: String -> UserInput -> String -> IO ()
psSPDIFile title spdi ps = 
  do { writeFile ps (psUserInput title spdi);
       putStrLn ("Generated ps file <"++ps++">") 
  } `catch` \err -> fail ("Cannot write to ps file <"++ps++">") 


initializeFile title npages = 
  unlines [ 
    "%!PS-Adobe-2.0",
    "%%Creator: SPDI-tool",
    "%%Title: ",
    "%%Pages: "++show npages,
    "%%PageOrder: Ascend",
    "%%BoundingBox: 0 0 "++
      show (xsize + 2*margin)++" "++ show (ysize+ 2*margin),
    "%%DocumentPaperSizes: A4",
    "%%EndComments",
    ""
  ]

initializePage scalingfactor minx miny =   
  unlines [
    translate margin margin, 
    setlinewidth 0 ++ newpath ++ moveto (-20) ++ rlineto vx ++ 
      rlineto vy ++ rlineto (-vx) ++ closepath ++ stroke,
    scale scalingfactor,
    translate (-minx) (-miny),
    setlinewidth 0,
    ""
  ] 
  where 
   vx = Vector (xsize+40, 0)
   vy = Vector (0, ysize+40)
   

terminatePage scalingfactor minx miny =
  scale (1 / scalingfactor) +++
  showpage

psUserInputSignature :: String -> UserInput -> [[(Vector)]] -> String
psUserInputSignature title userinput sig =
  initializeFile title 1 +++
  "%%Page: 1 1" +++
  initializePage scalingfactor cx cy +++
  drawEdges +++
  drawPath +++
  terminatePage scalingfactor cx cy +++
  "%%EOF"
  where
   vertices = vertexTable userinput
   vertex n = vertices !! n

   edges_of_region region =
      [(vertex s,vertex d) | s:d:_ <- tails (userRegionVertices region)]

   edges =
     concat $ map edges_of_region $ userRegions userinput

   ((minx, maxx), minLen, (miny, maxy)) = minmaxEdges edges

   xscale = xsize / (maxx - minx)
   yscale = ysize / (maxy - miny)
   scalingfactor = min xscale yscale

   cx = (maxx + minx - xsize/scalingfactor) / 2
   cy = (maxy + miny - ysize/scalingfactor) / 2

   drawEdges = unlines (map drawUserEdgePlain edges)

   arrow_size = minLen / 7

   sig'=concat sig
   drawPath = 
     (unlines $ map (psArrow arrow_size) $ zip sig' (safetail sig'))+++
     (unlines $ map (psArrow arrow_size) $ loopClosures sig)

   loopClosures [] = []
   loopClosures [r] = []
   loopClosures (r:s:rest) = (last s, head s):loopClosures rest 

   safetail [] = []
   safetail xs = tail xs

psUserInput :: String -> UserInput -> String
psUserInput title userinput =
  initializeFile title (1 + length (userRegions userinput)) +++
  firstPage +++
  concat' (zipWith drawRegion [1..] (userRegions userinput)) +++
  "%%EOF"
  where
   vertices = vertexTable userinput
   vertex n = vertices !! n

   edges_of_region region = 
      [(vertex s,vertex d) | s:d:_ <- tails (userRegionVertices region)]

   edges = 
     concat $ map edges_of_region $ userRegions userinput

   ((minx, maxx), minLen, (miny, maxy)) = minmaxEdges edges

   xscale = xsize / (maxx - minx)
   yscale = ysize / (maxy - miny)
   scalingfactor = min xscale yscale

   cx = (maxx + minx - xsize/scalingfactor) / 2
   cy = (maxy + miny - ysize/scalingfactor) / 2

   arrowHeadSize = minLen / 7
   drawEdges = unlines (map drawUserEdgePlain edges)
   labelPoints = 
     unlines (setfont (minLen/5):
       zipWith labelPoint (vertexTable userinput) (pointNames userinput)
     )
   labelPoint pos name = 
     showVector pos ++ " moveto (" ++ name ++ ") show stroke"

   firstPage = 
     "%%Page: 1 1" +++
     initializePage scalingfactor cx cy +++
     drawEdges +++
     labelPoints +++
     terminatePage scalingfactor cx cy
   drawRegion n region = 
     "%%Page: "++ show (n+1) ++ " " ++ show (n+1) +++
     setgray 0 +++ 
     setlinewidth 0 +++

     setfont (minLen*scalingfactor/5) +++
     drawVectors v_first v_second +++
     setfont 20 +++
     showPS (Vector (margin, margin+ysize-20)) ("Region "++show n) +++

     initializePage scalingfactor cx cy +++
     setgray 0.9 +++
     drawEdges +++

     setgray 0 +++
     setfont (minLen / 5) +++
     concat' (map drawLocalEdge edges)+++

     terminatePage scalingfactor minx miny
     where 
       parities = getParitiesOfRegion (vertexTable userinput) region

       vertices = userRegionVertices region
       v_first  = userRegionVFirst region
       v_second = userRegionVSecond region

       flip =
          let p1:p2:_ = vertices
              noflip  = (vertex p2 - vertex p1) >.< hat v_first > 0
              parity:_ = parities
          in if noflip then opposite parity else parity
       edges = zipWith 
                (\(s,d) p -> if p==flip then ((d,s),p) else ((s,d),p))
                (edges_of_region region) parities

       drawLocalEdge (edge, parity) =
          psArrow arrowHeadSize edge +++
          showVector midv ++ " moveto (" ++ show parity ++ ") show stroke"
          where 
            (v1,v2) = edge
            midv = (v1+v2) / constant' 2



drawVectors v1 v2 = 
  unlines(
   [ psArrow 0 (c, v1')
   , psArrow 0 (c, v2')
   , showVector (c + Vector (3,0)) ++ " moveto " ++
     showVector c ++ " 3 0 360 arc stroke"
   ] ++ 
   if v1 /= v2 then 
   [ showVector (c + (v1'-c) * constant 0.7) ++ " moveto (first) show stroke"
   , showVector (c + (v2'-c) * constant 0.5) ++ " moveto (second) show stroke"
   ] else [])
   where
     c   = Vector (144,144)
     v1' = c + 72 * v1 / constant' (modulus v1)
     v2' = c + 72 * v2 / constant' (modulus v2)

     theta1 = vectorToAngle v1
     theta2 = vectorToAngle v2

minmaxEdges [e] = minmaxEdge e
minmaxEdges (e:es) = 
  ( (min minx1 minx2, max maxx1 maxx2)
  , min l1 l2
  , (min miny1 miny2, max maxy1 maxy2)
  )
  where
    ((minx1, maxx1), l1, (miny1, maxy1)) = minmaxEdge  e
    ((minx2, maxx2), l2, (miny2, maxy2)) = minmaxEdges es

minmaxEdge (v1,v2) =
  ((min x1 x2, max x1 x2), l, (min y1 y2, max y1 y2))
  where 
    l = modulus (v2 - v1)
    (x1,y1) = fromVector v1
    (x2,y2) = fromVector v2

drawUserEdgePlain (v1,v2) =
  "newpath "++showVector v1++" moveto "++showVector v2++" lineto stroke"

psArrow :: RealPlus -> (Vector, Vector) -> String
psArrow len (v1,v2) = 
  "newpath " ++ showVector v1 ++ " moveto " ++
  showVector dv ++ " rlineto gsave " ++
  showVector vl ++ " rlineto stroke grestore "++
  showVector vr ++ " rlineto stroke"
  where 
    dv = v2 - v1
    v = if len > 0 
           then (constant' len * dv / constant' (modulus dv)) 
           else (dv / constant 7)
    vl = rotateVector 170 v
    vr = rotateVector (-170) v
    
showVector p = 
  show x ++ " " ++ show y
  where
    (x,y) = fromVector p


rlineto v = showVector v ++ " rlineto "
moveto v = showVector v ++ " moveto "
setlinewidth n = show n ++ " setlinewidth "
setfont size = 
  "/Times findfont "++show size++" scalefont setfont "
translate xshift yshift = 
  show xshift ++" "++show yshift ++" translate "
scale scalingfactor =
    show scalingfactor ++ " " ++ show scalingfactor ++ " scale "
setgray f = show f ++ " setgray "
showPS v s = moveto v ++ " ("++s++") show stroke "
showpage = "showpage "
stroke = "stroke "
newpath = "newpath "
closepath = "closepath "

line1 +++ line2 = line1 ++ "\n" ++ line2
concat' = foldl (+++) []

