module Parser (
  runProgram, runProgramFix, isNumber,
  parse, parseFix,
  parseAnyString, parseRealPlus, parseInteger, parseBoolean,
  parseFilename, parseFilenameExtension, 
  parseInterval, parseSignature, parseEdges, parseEdge, 
  parseSwitch, parseOption, 
  parseOption1Parameter, parseOption2Parameter, parseOption3Parameter,
  parseNamedOption1Parameter, parseNamedOption2Parameter, parseNamedOption3Parameter,
  parseOptional, defaultsTo, (<|>), (<+>), applyParser,
  parseEdgesFromFile, parseSignatureFromFile
) where

{-- 
Main function:

runProgram (program_welcome_string, program_usage_string) argumentParser
interprets the arguments either failing if it does not match or returning 
IO same structure as the parser if it succeeds.

Example usage of the parser, for the utility "simsig2fig" with the following commandline expected:
  simsig2fig [-over|-under] [-o file.fig] <file.spdi> <input interval> <exit interval> <signature>

     parse
         ( parseSwitch ["under", "over"]
         , parseOption1Parameter "o" parseFilename `defaultsTo` ""
         , parseFilename 
         , (parseInterval, parseInterval) 
         , parseSignature
         )
     

Notes:
 - parse takes a structure of parsers (using tuples and lists) and
   a string and returns a value Nothing if no parse is found, 
   Just (structure) if one is found.
 
 Parser constructors: Note that each parser has a return type. If p is a parser of type a, then
 (parse p) would return a value in (Maybe a).
 - parseFilename reads a filename. It returns a string.
 - parseFilenameExtension takes an extension and reads a filename. The extension is appended to the 
   filename if necessary.
 - parseOption takes a string and expects the given option (the string preceeded by a '-').
   It returns the string.
 - parseSwitch reads one of the given strings as an option - if missing it defaults to the first.
   It returns a string.
 - parseRealPlus reads a number an returns a RealPlus (Extend Double). Syntax accepted eg
   12, 12.02, .32, -1, +3.15, + 4, - 6.46, 1.02e12, 1.02e-12, - 1.02E +12 (in exponential form, the
   'e' must not be separated from the previous number - ie 1.2 e7 is not accepted).
 - parseAnyString reads a string, and also returns one.
 - parseBoolean reads a boolean value (where t, true, on, yes result in True, f, false, off,
   no result in False - case does not matter). It returns a boolean value.
 - parseInteger returns an Int type. Syntax allows a preceeding + or - sign.
 - parseInterval reads an interval (syntax eg [+21.0, 43]). Returns (RealPlus,RealPlus).
 - parseEdge reads an edge. Returns (String, String). Syntax is point-point (spaces allowed between
   point names and '-').
 - parseEdges reads an edgelist - a non-empty, comma separated list of edge names. 
   Returns [(String, String)].
 - parseSignature reads a signature, returns [ [(String, String)] ]. Syntax is a number of 
   (possibly empty) sequential edgelists with loop (non-empty) edgelists in between. Spaces
   are disregarded:
   
      edgelist(edgelist)*edgelist(edgelist)*edgelist...(edgelist)*edgelist
 
   We allow sequential edgelists also to be enclosed within brackets, and loops to be
   given as [edgelist] rather than (edgelist)* (for backward compatibility).
 
 - parser `defaultsTo` value tries to parse using the given parser, but defaults to 
   the given value if not given. Note that the type returned by the parser and the value
   must match.
 - parseOption1Parameter takes the name of an option (which takes one additional parameter)
   and the parser to use for the parameter. It returns whatever the type the parser returns.
   Similarly, parseOption2Parameter to parseOption3Parameter take 2 and 3 parameter parsers
   and return a pair, or triple of the expected types.
 - parseNamedOption1Parameter is the same as parseOption1Parameter but returns a pair with the
   name of the option appearing as the first of the pair. Similarly parseNamedOption2Parameter
   and parseNamedOption3Parameter.
  
 - parseOptional parser: if parse returns an object of type a, this returns an object of type Maybe a.
   The value is Nothing if it left out, Just x if object x is found
 - (parser1 <|> parser2) tries parser1 on the input, and if it fails, tries parser2. This can be
   used to allow options such as: <signature> | -sig <signature>:
   
      (parseSignature <|> parseNamedOption1Parameter "sig" parseSignature)

   Note that the two parsers must be of the same type.
 - (parser1 <+> parser2) is the same as <|>, but for different typed parsers. If parser1 returns
   type a and parser2 returns type b, (parser1 <+> parser2) is of return type (Maybe a, Maybe b).
   For example: <signature> | -sig <signature-file>:
       
      (parseNamedOption1Parameter "sig" parseFilename <+> parseSignature)

   If a -sig option is found, the parse will return (Just filename, Nothing), if not,
   but a signature is correctly parsed, it returns (Nothing, Just signature). If neither
   is correctly parsed, the parsing would fail.
 - applyParser f parser, applies function f to the result of the parser:

       applyParser (> 10) parseRealPlus


runProgramFix takes a fix-point function for the parser, rather than the parser itself. It is e
easiest to see how it works by seeing the underlying parseFix. If parts of the parser may depend on 
the values other parts of the parser return, parseFix greatly simplify the parser description. 
eg [-pascal | -c] infile can be written as:
   parseFix (\(language, _) -> (parseOption "c" <|> parseOption "pascal", parseFilenameExtension language))

   [-o outfile] infile, where the output file default value is that of the input file, but with changed
   extension:

   parseFix (\(_, input) -> 
      ( parseOption1Parameter "o" parseFilename `defaultsTo` changeExtension input
      , parseFilename
      )
   )

Note that parseFix takes a function, from the parse result to the parser. It is up to the programmer
to make sure that this does not lead to circular dependancies.
 
--}

import Char
import List
import Maybe
import System

import Infinity
import Interval

infixr 2 <+>, <|>
infixl 3 <*>
infixr 4 <*, *>

runProgram :: Parse a b => (String, String) -> a -> IO b
runProgram (program_description, program_usage) parameter_parser =
  do 
    putStrLn program_description
    arguments <- getArgs
    case parse parameter_parser (unwords arguments) of
      Nothing         -> fail program_usage 
      Just parameters -> return parameters

runProgramFix :: Parse a b => (String, String) -> (b -> a) -> IO b
runProgramFix (program_description, program_usage) parameter_parser =
  do 
    putStrLn program_description
    arguments <- getArgs
    case parseFix parameter_parser (unwords arguments) of
      Nothing         -> fail program_usage
      Just parameters -> return parameters


type Parser a = String -> Maybe (a, String)

parseEOS :: Parser ()
parseEOS str 
 | all isSpace str = return ((), "")
 | otherwise       = Nothing

constantParser :: a -> Parser a
constantParser x = \str -> return (x,str)

applyParser :: (a -> b) -> Parser a -> Parser b
applyParser f parser str =
  do { (x, rest) <- parser str; return (f x, rest) }

(<*>) :: Parser a -> Parser b -> Parser (a,b)
(p1 <*> p2) str =
   do { (x, rest) <- p1 str; (x', rest') <- p2 rest; return ((x,x'), rest') }

(*>) :: Parser a -> Parser b -> Parser b
(p1 *> p2) str = do { ((x,y), rest) <- (p1 <*> p2) str ; return (y, rest) }

(<*) :: Parser a -> Parser b -> Parser a
(p1 <* p2) str = do { ((x,y), rest) <- (p1 <*> p2) str ; return (x, rest) }

parseList :: [Parser a] -> Parser [a]
parseList [] = constantParser []
parseList (p:ps) = applyParser (uncurry (:)) (p <*> parseList ps)

(<-<) :: (b -> Parser a) -> Parser b -> Parser (a,b)
(fp1 <-< p2) input = 
  let result@ ~(Just ((_,v), _)) = (fp1 v <*> p2) input
  in  result

(<|>) :: Parser a -> Parser a -> Parser a 
(p1 <|> p2) str =
  let result = p1 str 
  in  if isNothing result then p2 str else result
  
(<+>) :: Parser a -> Parser b -> Parser (Maybe a, Maybe b)
p1 <+> p2 = 
  let maybe_p1 = applyParser Just p1
      maybe_p2 = applyParser Just p2
      nothing  = constantParser Nothing
  in  (maybe_p1 <*> nothing) <|> (nothing <*> maybe_p2)

(<^>) :: Parser a -> Int -> Parser [a]
(parser <^> 0) str = return ([], str)
(parser <^> n) str =
  do { (x, rest) <- parser str; (xs, rest') <- (parser <^> (n-1)) rest; return (x:xs, str) }

parseAny :: [Parser a] -> Parser a
parseAny [parser] = parser 
parseAny (p:ps) = p <|> parseAny ps

parseStar :: Parser a -> Parser [a]
parseStar parser = parsePlus parser <|> constantParser []

parsePlus :: Parser a -> Parser [a]
parsePlus parser = applyParser (uncurry (:)) (parser <*> parseStar parser)

--
parseChar' :: Char -> Parser Char
parseChar' c = parseCharProp' (c==)

parseCharProp' :: (Char -> Bool) -> Parser Char
parseCharProp' cond cs =
  case cs of
    (c:cs') -> if cond c then return (c, cs') else Nothing
    _       -> Nothing

parseWord' :: String -> Parser String
parseWord' word = parseList (map parseChar' word) 

parseWhiteSpace = parseStar (parseCharProp' isSpace)
parseNonEmptyWhiteSpace = parsePlus (parseCharProp' isSpace)
--
parseChar :: Char -> Parser Char
parseChar c = parseCharProp (c==)

parseCharProp :: (Char -> Bool) -> Parser Char
parseCharProp cond cs =
  case cs of
    (c:cs') -> if cond c then return (c, dropWhile isSpace  cs') else Nothing
    _       -> Nothing

parseAnyWord :: Parser String
parseAnyWord = parseWordProp (\word -> word/="" && head word /= '-')

parseWord :: String -> Parser String
parseWord word = parseWordProp (word ==)

parseWordProp :: (String -> Bool) -> Parser String
parseWordProp expected str
  | word == "" || not (expected word) = Nothing
  | otherwise                         = return (word, dropWhile isSpace rest)
  where
    (word, rest) = break isSpace str


parseBoolean :: Parser Bool
parseBoolean =
  applyParser (const True) (parseWordProp ((`elem` ["T","TRUE", "ON", "YES"]) . map toUpper)) <|>
  applyParser (const False)(parseWordProp ((`elem` ["F","FALSE","OFF","NO" ]) . map toUpper))

parseInteger :: Parser Int
parseInteger =
  applyParser (read . concat) $ 
    parseList 
      [ parseWord' "-" <|> applyParser (const "") (parseWord' "+") <|> constantParser ""
      , applyParser (const "") parseWhiteSpace
      , parsePlus (parseCharProp' isDigit)
      ]

isNumber str = 
  case (parseWhiteSpace *> parseRealPlus <* parseWhiteSpace) str of
    Just (n, "") -> Just n
    _            -> Nothing

parseRealPlus :: Parser (RealPlus)
parseRealPlus = 
  applyParser read $ 
    ((applyParser concat $
       parseList 
         [ parseSignedRational
         , (parseWord' "e" <|> parseWord' "E") <* parseWhiteSpace 
         , parseOptionalSign
         , parseDigits
         ]
      ) <|> parseSignedRational
    ) <* parseWhiteSpace
  where
    parseDigits   = parsePlus (parseCharProp' isDigit)
    parseRational = applyParser concat (parseList [parseDigits <|> constantParser "0", parseWord' ".", parseDigits]) 
                 <|> parseDigits
    parseSignedRational = applyParser (uncurry (++)) (parseOptionalSign <*> parseRational)
    parseOptionalSign = (parseWord' "-" <|> applyParser (const "") (parseWord' "+") <|> constantParser "") <* parseWhiteSpace
   
    
   

parseNonOption :: Parser String
parseNonOption = parseWordProp (not . isOption)
  where
    isOption str = (str /= "") && (head str == '-')

-- parsers for command line in speedi tools

parseOptional :: Parser a -> Parser (Maybe a)
parseOptional parser = 
  applyParser Just parser `defaultsTo` Nothing

defaultsTo :: Parser a -> a -> Parser a
parser `defaultsTo` value = 
  parser <|> constantParser value

parseSwitch :: [String] -> Parser String
parseSwitch options@(value:_) = 
  parseAny (map parseOption options) <|> constantParser value

parseOption :: String -> Parser String
parseOption option = applyParser tail $ parseWord ('-':option)

parseFullOption :: String -> Int -> Parser [String]
parseFullOption option parameter_count = 
  parseOption option *> (parseNonOption <^> parameter_count)

parseOption1Parameter option parser1 = 
   applyParser snd (parseNamedOption1Parameter option parser1)
parseOption2Parameter option parser1 parser2 = 
   applyParser snd (parseNamedOption2Parameter option parser1 parser2)
parseOption3Parameter option parser1 parser2 parser3 = 
   applyParser snd (parseNamedOption3Parameter option parser1 parser2 parser3)

parseNamedOption1Parameter option parser1 = 
   parseOption option <*> parser1 
parseNamedOption2Parameter option parser1 parser2 =
   parseOption option <*> (parser1 <*> parser2)
parseNamedOption3Parameter option parser1 parser2 parser3 =
   parseOption option <*> (applyParser (\((x,y),z)->(x,y,z)) (parser1 <*> parser2 <*> parser3))


   
parseInterval :: Parser (Interval RealPlus)
parseInterval = applyParser toClosedInterval $
   parseChar '['  *> parseRealPlus <* parseChar ',' <*> parseRealPlus <* parseChar ']'

parsePoint :: Parser String
parsePoint = parsePlus (parseCharProp' (`notElem` "-\n \t()!?,")) <* parseWhiteSpace

parseEdge :: Parser (String, String)
parseEdge = 
  (parsePoint <* parseChar' '-' <* parseWhiteSpace) <*> parsePoint <* parseWhiteSpace

parseEdges :: Parser [(String, String)]
parseEdges = applyParser (uncurry (:)) 
       (parseEdge <*> parseStar (parseChar' ',' *> parseWhiteSpace *> parseEdge))

parseSignature :: Parser [[(String,String)]]
parseSignature = parseSignatureNew <|> parseSignatureOld

parseSignatureOld = 
  applyParser (\(xss,xs) -> concat xss ++ [xs])
      (parseStar parseSignatureRS <*> parseSignatureR)
  where
    parseEdges'     = parseEdges <|> constantParser []
    parseSignatureR  = parseEdges'
    parseSignatureRS =  parseList [parseChar '[' *> parseEdges' <* parseChar ']']
                    <|> parseList [parseEdges', parseChar '[' *> parseEdges' <* parseChar ']']

parseSignatureNew = 
  applyParser (\(ess,es) -> concat ess ++ [es])
      (parseStar parseSignatureRS <*> parseSignatureR)
  where
    parseEdges'     = parseEdges <|> constantParser []
    parseSignatureR  =  (parseChar '(' *> parseEdges' <* parseChar ')') <|> parseEdges'
    parseSignatureRS =  parseList [ parseEdges'
                                  , parseChar '(' *> parseEdges' <* parseChar ')' <* parseChar '*'
                                  ]
                    <|> parseList [ parseChar '(' *> parseEdges' <* parseChar ')'
                                  , parseChar '(' *> parseEdges' <* parseChar ')' <* parseChar '*'
                                  ]

parseFilename :: Parser String
parseFilename = parseAnyWord

parseFilenameExtension :: String -> Parser String
parseFilenameExtension extension = applyParser addExtension parseFilename
  where
    addExtension filename = 
      -- TEST whether finishes with .extension:
      -- reverse ('.':extension) `isPrefixOf` reverse filename 
      if '.' `elem` filename
        then filename 
        else filename ++ '.':extension

parseAnyString = parseAnyWord

-- Parsing tuples

parse :: Parse a b => a -> String -> Maybe b
parse parser input = do 
  { (result, _) <- 
       (parseWhiteSpace *> wrapParser parser <* parseEOS) input
  ; return result
  }

parseFix :: Parse a b => (b -> a) -> String -> Maybe b
parseFix fparser input = 
  let result@ ~(Just value) = parse (fparser value) input 
  in  result

class Parse a b where
  wrapParser :: a -> Parser b

instance Parse (Parser a) a where
  wrapParser parser = parser

instance Parse a b => Parse [a] [b] where
  wrapParser xs = parseList (map wrapParser xs)

instance (Parse a1 b1, Parse a2 b2) => Parse (a1,a2) (b1,b2) where
  wrapParser (x1,x2) = wrapParser x1 <*> wrapParser x2

instance (Parse a1 b1, Parse a2 b2, Parse a3 b3) => 
  Parse (a1,a2,a3) (b1,b2,b3) where
  wrapParser (x1,x2,x3) = 
       applyParser (\((y1,y2),y3) -> (y1,y2,y3)) $
		 (wrapParser x1<*>wrapParser x2<*>wrapParser x3)

instance (Parse a1 b1, Parse a2 b2, Parse a3 b3, Parse a4 b4) => 
  Parse (a1,a2,a3,a4) (b1,b2,b3,b4) where
  wrapParser (x1,x2,x3,x4) = 
       applyParser (\(((y1,y2),y3),y4) -> (y1,y2,y3,y4)) $
       (wrapParser x1<*>wrapParser x2<*>wrapParser x3<*>wrapParser x4)

instance (Parse a1 b1, Parse a2 b2, Parse a3 b3, Parse a4 b4, Parse a5 b5) => 
  Parse (a1,a2,a3,a4,a5) (b1,b2,b3,b4,b5) where
  wrapParser (x1,x2,x3,x4,x5) = 
       applyParser (\((((y1,y2),y3),y4),y5) -> (y1,y2,y3,y4,y5)) $
       (wrapParser x1<*>wrapParser x2<*>wrapParser x3<*>wrapParser x4<*>
        wrapParser x5)

instance 
  (Parse a1 b1, Parse a2 b2, Parse a3 b3, Parse a4 b4, Parse a5 b5
  ,Parse a6 b6) =>
  Parse (a1,a2,a3,a4,a5,a6) (b1,b2,b3,b4,b5,b6) where
  wrapParser (x1,x2,x3,x4,x5,x6) = 
       applyParser 
       (\(((((y1,y2),y3),y4),y5),y6) -> (y1,y2,y3,y4,y5,y6)) $
       (wrapParser x1<*>wrapParser x2<*>wrapParser x3<*>wrapParser x4<*>
        wrapParser x5<*>wrapParser x6)

instance 
  (Parse a1 b1, Parse a2 b2, Parse a3 b3, Parse a4 b4, Parse a5 b5
  ,Parse a6 b6, Parse a7 b7) =>
  Parse (a1,a2,a3,a4,a5,a6,a7) (b1,b2,b3,b4,b5,b6,b7) where
  wrapParser (x1,x2,x3,x4,x5,x6,x7) = 
       applyParser 
       (\((((((y1,y2),y3),y4),y5),y6),y7) -> (y1,y2,y3,y4,y5,y6,y7)) $
       (wrapParser x1<*>wrapParser x2<*>wrapParser x3<*>wrapParser x4<*>
        wrapParser x5<*>wrapParser x6<*>wrapParser x7)

instance 
  (Parse a1 b1, Parse a2 b2, Parse a3 b3, Parse a4 b4, Parse a5 b5
  ,Parse a6 b6, Parse a7 b7, Parse a8 b8) =>
  Parse (a1,a2,a3,a4,a5,a6,a7,a8) (b1,b2,b3,b4,b5,b6,b7,b8) where
  wrapParser (x1,x2,x3,x4,x5,x6,x7,x8) = 
       applyParser 
       (\(((((((y1,y2),y3),y4),y5),y6),y7),y8) -> 
                                            (y1,y2,y3,y4,y5,y6,y7,y8)) $
       (wrapParser x1<*>wrapParser x2<*>wrapParser x3<*>wrapParser x4<*>
        wrapParser x5<*>wrapParser x6<*>wrapParser x7<*>wrapParser x8)

instance 
  (Parse a1 b1, Parse a2 b2, Parse a3 b3, Parse a4 b4, Parse a5 b5
  ,Parse a6 b6, Parse a7 b7, Parse a8 b8, Parse a9 b9) =>
  Parse (a1,a2,a3,a4,a5,a6,a7,a8,a9) (b1,b2,b3,b4,b5,b6,b7,b8,b9) where
  wrapParser (x1,x2,x3,x4,x5,x6,x7,x8,x9) = 
       applyParser 
       (\((((((((y1,y2),y3),y4),y5),y6),y7),y8),y9) -> 
                                         (y1,y2,y3,y4,y5,y6,y7,y8,y9)) $
       (wrapParser x1<*>wrapParser x2<*>wrapParser x3<*>wrapParser x4<*>
        wrapParser x5<*>wrapParser x6<*>wrapParser x7<*>wrapParser x8<*>
        wrapParser x9)

instance 
  (Parse a1 b1, Parse a2 b2, Parse a3 b3, Parse a4 b4, Parse a5 b5
  ,Parse a6 b6, Parse a7 b7, Parse a8 b8, Parse a9 b9, Parse a0 b0) =>
  Parse (a1,a2,a3,a4,a5,a6,a7,a8,a9,a0) (b1,b2,b3,b4,b5,b6,b7,b8,b9,b0) where
  wrapParser (x1,x2,x3,x4,x5,x6,x7,x8,x9,x0) = 
       applyParser 
       (\(((((((((y1,y2),y3),y4),y5),y6),y7),y8),y9),y0) -> 
                                      (y1,y2,y3,y4,y5,y6,y7,y8,y9,y0)) $
       (wrapParser x1<*>wrapParser x2<*>wrapParser x3<*>wrapParser x4<*>
        wrapParser x5<*>wrapParser x6<*>wrapParser x7<*>wrapParser x8<*>
        wrapParser x9<*>wrapParser x0)


--

parseFile:: String -> Parser a -> String -> IO a
parseFile filename parser error_string = 
  do { text <- readFile filename `catch` (\_ -> fail ("Cannot read file <"++filename++">"))
     ; case parse parser (filter (`notElem` "\t\n") text) of
         Nothing -> fail error_string
         Just v  -> return v
     }

parseEdgesFromFile filename = 
  parseFile filename parseEdges 
    ("Cannot parse edge-list in <"++filename++">")

parseSignatureFromFile filename = 
  parseFile filename parseSignature
    ("Cannot parse signature in <"++filename++">")




  
