module Signature (
  Path(..), Signature(..),
  toSignature, fromSignature, toPath, fromPath,
  lastPath, headPath, tailPath, initPath, (!++), (!:), 
  lastSignature, headSignature, tailSignature, initSignature, (!!++), (!!:),
  isLoop,

  showSignature, showPath, showEdge,
  showStringSignature, showStringPath, showStringEdge,

  readSignature, readPath, readEdge,
  readUnambiguousEdge,
) where

import Maybe
import List

import Interval
import Graph

newtype Path = Path [EdgeId] deriving (Eq, Show)
newtype Signature = Signature [Path] deriving (Eq, Show)

-- 'Constructors'
toSignature :: [Path] -> Signature
toSignature = Signature

toPath :: [EdgeId] -> Path
toPath = Path

-- 'Destructors'
fromSignature :: Signature -> [Path]
fromSignature (Signature ws) = ws

fromPath :: Path -> [EdgeId]
fromPath (Path es) = es

-- Operations
lastPath, headPath :: Path -> EdgeId
headPath = head . fromPath
lastPath = last . fromPath

initPath, tailPath :: Path -> Path
tailPath = toPath . tail . fromPath
initPath = toPath . init . fromPath

lastSignature, headSignature :: Signature -> Path
headSignature = head . fromSignature
lastSignature = last . fromSignature

initSignature, tailSignature :: Signature -> Signature
tailSignature = toSignature . tail . fromSignature
initSignature = toSignature . init . fromSignature

(!++) :: Path -> Path -> Path
w1 !++ w2 = toPath (fromPath w1 ++ fromPath w2)

(!:) :: EdgeId -> Path -> Path
e !: w = toPath (e : fromPath w)

(!!++) :: Signature -> Signature -> Signature
s1 !!++ s2 = toSignature (fromSignature s1 ++ fromSignature s2)

(!!:) :: Path -> Signature -> Signature
p !!: s = toSignature ( p: fromSignature s)

isLoop :: Signature -> Bool
isLoop signature =
  (length (fromSignature signature) >= 2) &&
  ([headPath $ lastSignature $ initSignature $ signature] == 
   (fromPath $ lastSignature $ signature))

-- The following code enables us to show a signature/path/edge 
-- in a user-friendly way, by using translation functions 
-- to name points and edges as expected.

showEdge :: (EdgeId -> (String, String)) -> EdgeId -> String
showEdge edgeID2points eid = showStringEdge (edgeID2points eid)

showPath :: (EdgeId -> (String, String)) -> Path -> String
showPath edgeID2points path = 
  showStringPath $ map edgeID2points $ fromPath path

showSignature :: (EdgeId -> (String, String)) -> Signature -> String
showSignature edgeID2points signature =
  showStringSignature $  map (map edgeID2points . fromPath) $ fromSignature signature


showStringEdge (src,dst) = src++"-"++dst

showStringPath = concat . intersperse ", " . map showStringEdge 

showStringSignature [] = ""
showStringSignature [r] = '(':showStringPath r++")"
showStringSignature (r:s:sig) = '(':showStringPath r++")("++showStringPath s++")*"++showStringSignature sig

-- 
-- Parse:
--   - pairs of strings into edges, 
--   - lists of pairs of strings into Paths, 
--   - lists of lists of pairs of strings into Signatures
--

readUnambiguousEdge (point2vector, points2edgeID) (v1,v2)
  | isNothing (point2vector v1) = fail ("Point <"++v1++"> undefined")
  | isNothing (point2vector v2) = fail ("Point <"++v2++"> undefined")
  | otherwise = case points2edgeID (v1,v2) of
                  []    -> fail ("There is no edge corresponding to "++v1++"-"++v2)
                  [eid] -> return eid
                  _     -> fail ("Edge "++v1++"-"++v2++" is ambiguous")

readEdge (point2vector,points2edgeID) (v1,v2) 
  | isNothing (point2vector v1) = fail ("Point <"++v1++"> undefined")
  | isNothing (point2vector v2) = fail ("Point <"++v2++"> undefined")
  | edges == []                 = fail ("There is no edge corresponding to "++v1++"-"++v2)
  | otherwise                   = return eid
  where
    edges@ ~(eid:_) = points2edgeID (v1,v2)

readPath _ [] = return (toPath [])
readPath inputs (edge:edgelist) = 
  do { eid <- readEdge inputs edge; path <- readPath inputs edgelist; return (eid !: path) }

readSignature inputs signature 
  | null signature     = fail "A signature may not be empty"
  | any null loops     = fail "Loops in a signature may not be empty"
  | null last_sequence = fail "Last sequential path signature may not be empty"
  | otherwise          = readSignature' inputs signature
  where
    loops = [ xs | (0, xs) <- zip [n `mod` 2 | n <- [1..]] signature ]
    last_sequence = last signature
    
readSignature' _ [] = return (toSignature [])
readSignature' inputs (path:signature) =
  do { path' <- readPath inputs path
     ; signature' <- readSignature' inputs signature
     ; return (path' !!: signature')
     }
