module Successor (
  LoopType (..), loopType,
  isStay, isExitLeft, isExitBoth, isExitRight, isDie,
  reach, reach1, feasibleLoop, exitLoop1, applySignature,
  fixpoints
) where

import List 
import Maybe 
import Graph

import Signature
import Interval
import AffineFunctions

data LoopType =
    Stay (RealPlus, RealPlus) 
  | ExitLeft RealPlus 
  | ExitRight RealPlus 
  | ExitBoth 
  | Die

isStay, isExitLeft, isExitBoth, isExitRight, isDie:: LoopType -> Bool
isStay (Stay _) = True
isStay _ = False

isExitLeft (ExitLeft _) = True
isExitLeft _ = False

isExitBoth ExitBoth = True
isExitBoth _ = False

isExitRight (ExitRight _) = True
isExitRight _ = False

isDie Die = True
isDie _ = False

instance Show LoopType where
  show looptype =
    case looptype of
      ExitLeft  fpr  -> "Exit left: right fixed point is "++show fpr 
      ExitRight fpl  -> "Exit right: left fixed point is "++show fpl 
      ExitBoth       -> "Exit both"
      Die            -> "Die"
      Stay (fpl,fpr) -> "Stay: fixed points are ("++show fpl++", "++show fpr++")"

-- Increasing is a pair of boolean values representing whether the values of a given function over intervals are increasing or decreasing 
type Increasing = (Bool, Bool) 

-- Given an interval and the Successor of this intervals, it returns whether the each extremes of the interval are increasing or not
increasing :: Interval RealPlus -> Interval RealPlus -> Increasing
increasing i suc_i = (left_i < left_suc_i, right_i < right_suc_i)
 where
    (l_i, r_i) = fromInterval i
    (left_i, right_i) = (value l_i, value r_i)
    (l_si, r_si) = fromInterval suc_i
    (left_suc_i, right_suc_i) = (value l_si, value r_si)


-- Apply a signature to an interval
applySignature :: 
  Graph -> Interval RealPlus -> Signature -> [Interval RealPlus]
applySignature graph interval signature = 
  let paths    = fromSignature signature

      exit_paths = map concat (tails (map fromPath paths)) -- avoid problems with empty r's
      initials = tail (map head exit_paths)

  in  alternate [interval] (zip paths initials)
  where
    alternate intervals0 [(r,_)] = map (apply (multiStepComp graph r)) intervals0
    alternate intervals0 ((r,head_s):(s,head_r):rest) = 
      let intervals1 = exitSeq  graph r head_s intervals0
          intervals2 = exitLoop graph s head_r intervals1
      in  alternate intervals2 rest
 
-- Given a path (sequence of edges) of type r, and the head of the
-- next path (s), and a starting interval, calculate the resulting
-- intervals (a set -list - of intervals)
exitSeq1 :: Graph -> Path -> EdgeId -> Interval RealPlus -> [Interval RealPlus]
exitSeq1 graph r head_s interval = 
  let tmaf = multiStepComp graph (r !++ toPath [head_s])
  in  [apply tmaf interval]

-- Apply the above function to a set (list) of input intervals
exitSeq :: Graph -> Path -> EdgeId -> [Interval RealPlus] -> [Interval RealPlus]
exitSeq graph path edgeId = concat . (map (exitSeq1 graph path edgeId))


-- Is a loop feasible?
feasibleLoop :: Graph -> EdgeId -> Path -> Bool
feasibleLoop graph exit loop = 
  not (all isEmpty $  
    exitLoop1 graph loop exit (toInterval (closed InfiNeg, closed InfiPos))
  )



-- Given a path (sequence of edges) of type s, and the head of the
-- next path (r), and a starting interval, calculate the resulting
-- intervals (a set -list - of intervals)
exitLoop1 :: Graph -> Path -> EdgeId -> Interval RealPlus -> [Interval RealPlus]
-- 
exitLoop1 graph loop head_r interval
     | isEmpty sIntersectLoop             = [apply exit_tmaf appliedLoop]
     | (extr_L <= fpl) && (fpu <= extr_U) = -- STAY
         if head_r == headPath loop then [extr] else [emptyInt]
     | (fpl > extr_U) || (fpu < extr_L)   = exit_die    -- DIE
     | otherwise                          = [appliedLU] 
       -- EXIT-BOTH, EXIT-RIGHT, EXIT-LEFT.
    where 
      tmaf0 = multiStepComp graph loop
      last_s = lastPath loop
      exit_tmaf = oneStepComp graph tmaf0 (last_s, head_r)
      loop_tmaf = oneStepComp graph tmaf0 (last_s, headPath loop)
      (s,j) = fromTMAF2SJ loop_tmaf 
      int_LU = s /\ j
      (extr_L, extr_U) = values int_LU

      extr = fixpoints (fromTMAF2MAF loop_tmaf) interval (edgeRange graph $ headPath loop)
      (extr_l, extr_u) = fromInterval extr
      (fpl, fpu) = (value extr_l, value extr_u)

      appliedLoop = apply loop_tmaf interval
      appliedLU = apply exit_tmaf (apply loop_tmaf int_LU)
      sIntersectLoop = appliedLoop /\ s 
         -- gives the iterable interval: i \in s/\j
      exit_die = exitDie sIntersectLoop loop_tmaf exit_tmaf


-- Apply the above function 'exitLoop1' to a list of intervals
exitLoop :: Graph -> Path -> EdgeId -> [Interval RealPlus] -> [Interval RealPlus]
exitLoop graph path edgeId = concat . (map (exitLoop1 graph path edgeId))


-- Function that compute the set (list) of exit intervals for the DIE case
exitDie :: Interval RealPlus -> TMAF -> TMAF -> [Interval RealPlus]
exitDie interval tmaf_loop tmaf_exit
  | isEmpty interval = []
  | otherwise        = exit_interval:exitDie interval_succ tmaf_loop tmaf_exit
  where
    interval_succ = apply tmaf_loop interval
    exit_interval = apply tmaf_exit interval_succ


test1 :: Graph -> Path -> Interval RealPlus -> Interval RealPlus -> Bool
-- test1 graph loop final_interval intervals_to_test increasing_interval
test1 graph loop i_f i_0
     | isEmpty sIntersectLoop               = False
     | (extr_L <= fp_l) && (fp_u <= extr_U) = test_stay i_0 loop_tmaf i_f i_fp inc -- STAY
     | (fp_l > extr_U) || (fp_u < extr_L)   = test_die i_0 loop_tmaf i_f i_fp inc  -- DIE
     | (fp_l < extr_L) && (fp_u > extr_U)   = test_both loop_tmaf i_f int_LU -- BOTH
     | (extr_L <= fp_l) && (extr_U < fp_u)  = test_right i_0 loop_tmaf i_f i_fp int_LU inc -- RIGHT
     |  (fp_l < extr_L) && (fp_u <= extr_U) = test_left i_0 loop_tmaf i_f i_fp int_LU inc -- LEFT
    where
      tmaf0 = multiStepComp graph loop
      last_s = lastPath loop

      loop_tmaf = oneStepComp graph tmaf0 (last_s, headPath loop)

      (s,j) = fromTMAF2SJ loop_tmaf 

      int_LU = s /\ j
      (extr_L, extr_U) = values int_LU

      i_fp = fixpoints (fromTMAF2MAF loop_tmaf) i_0 (edgeRange graph $ headPath loop)
      (fp_l, fp_u) = values i_fp

      appliedLoop = apply loop_tmaf i_0
      sIntersectLoop = appliedLoop /\ s 

      inc = increasing i_0 appliedLoop

-- Test function: test whether the final point belongs to the interval given 
-- by the Exit algorithm when applied to a signature
test :: Graph -> Path -> Interval RealPlus -> [Interval RealPlus] -> Bool
test graph loop xf = or . (map (test1 graph loop xf))

test_stay :: Interval RealPlus -> TMAF -> Interval RealPlus -> Interval RealPlus -> Increasing -> Bool
-- test_stay interval loop_tmaf final_point Fixpoint_interval increasing_int
test_stay i_0 tmaf i_f i_fp increasing
  -- Not needed since fixed point interval is always open
  -- | (f_l == fp_u) || (fp_l == f_u)  = False 
  | i_f `overlaps` i_fp         = True -- overlaps rather than subset
  | (f_u <= fp_l) && not incr_l = False 
  | (f_l >= fp_u) && incr_u     = False 
  | otherwise                   = search i_0 tmaf i_f increasing
  where     
    (fp_l, fp_u) = values i_fp 
    (f_l, f_u) = values i_f 
    (incr_l, incr_u) = increasing


test_die :: Interval RealPlus -> TMAF -> Interval RealPlus -> Interval RealPlus -> Increasing -> Bool
-- test_die  interval loop_tmaf final_point Fixpoint_interval increasing_int
test_die i_0 loop_tmaf i_f i_fp increasing = 
  search i_0 loop_tmaf i_f increasing


test_both :: TMAF -> Interval RealPlus -> Interval RealPlus -> Bool
-- test_both loop_tmaf final_point [L,U] 
test_both loop_tmaf i_f int_LU 
   = i_f `overlaps` appliedLU 
    where
      appliedLU = apply loop_tmaf int_LU

test_left :: Interval RealPlus -> TMAF -> Interval RealPlus -> Interval RealPlus -> Interval RealPlus -> Increasing -> Bool
-- test_left  i_0 loop_tmaf final_point Fixpoint_i_0 [L,U] increasing_int
test_left i_0 loop_tmaf i_f i_fp int_LU increasing 
   | i_f `overlaps` applied_L_fpu = True 
   | f_u < l_L_fpu                = False
   | (u_L_fpu < f_l) && incr_u    = False
   | otherwise                    = search i_0 loop_tmaf i_f increasing
 where
    (f_l, f_u) = values i_f
    (extr_fp_l, extr_fp_u) = fromInterval i_fp
    (fp_l, fp_u) = (value extr_fp_l, value extr_fp_u)
    (incr_l, incr_u) = increasing
    (extrL, extrU) = fromInterval int_LU
    int_Lfpu = toInterval (extrL, extr_fp_u)
    applied_L_fpu = apply loop_tmaf int_Lfpu
    (l_Lfpu, u_Lfpu) = fromInterval applied_L_fpu
    (l_L_fpu, u_L_fpu) = (value l_Lfpu, value u_Lfpu)


test_right :: Interval RealPlus -> TMAF -> Interval RealPlus -> Interval RealPlus -> Interval RealPlus -> Increasing -> Bool
-- test_right  i_0 loop_tmaf final_point Fixpoint_i_0 [L,U] increasing_int
test_right i_0 loop_tmaf i_f i_fp int_LU increasing 
   | i_f `overlaps` applied_fpl_U  = True 
   | f_l > fpl_U_u                 = False 
   | (f_u < fpl_U_l) && not incr_l = False 
   | otherwise                     = search i_0 loop_tmaf i_f increasing
 where
    (f_l, f_u)             = values i_f
    (extr_fp_l, extr_fp_u) = fromInterval i_fp
    (fp_l, fp_u)           = (value extr_fp_l, value extr_fp_u)
    (incr_l, incr_u)       = increasing
    (extrL, extrU)         = fromInterval int_LU
    int_fplU               = toInterval (extr_fp_l, extrU)
    applied_fpl_U          = apply loop_tmaf int_fplU
    (fplU_l, fplU_u)       = fromInterval applied_fpl_U
    (fpl_U_l, fpl_U_u)     = (value fplU_l, value fplU_u)

-- Search: 
search :: Interval RealPlus -> TMAF -> Interval RealPlus -> Increasing -> Bool
search i_0 loop_tmaf i_f increasing
   | i_f `overlaps` i_0          = True
   | isEmpty i_0                 = False
   | (f_u < i_0_l) && incr_l     = False
   | (f_l > i_0_u) && not incr_u = False
   | succ_int == i_0             = False  -- not sure whether it is necessary
   | otherwise                   = search succ_int loop_tmaf i_f increasing
  where
    succ_int         = apply loop_tmaf i_0
    (f_l, f_u)       = values i_f
    (incr_l, incr_u) = increasing
    (i_0_l, i_0_u)   = values i_0

reach1 :: Graph -> Interval RealPlus -> Interval RealPlus ->
  Signature -> Bool
reach1 graph i_0 i_f signature 
  | isLoop signature  = 
         test graph penultimate i_f
            (applySignature graph i_0 signature2)
  | otherwise = satisfies (applySignature graph i_0 signature)
  where
    satisfies is = not (and ( map (isEmpty . (i_f /\)) is ))

    -- signature = signature2 :< penultimate :< ultimate
    signature1 = initSignature signature
    signature2 = initSignature signature1
 
    ultimate = lastSignature signature
    penultimate = lastSignature signature1

-- MAIN ALGORITHM
reach :: Graph -> Interval RealPlus -> Interval RealPlus -> 
         [Signature] -> [Path] -> [(Signature,Bool)]
reach graph i_0 i_f signatures loops = satisfying
-- signatures = non loop signatures
-- loops = final loops (starting and finishing on the destinations)
  where
     -- list of collections of intervals after applying signature 
     iss = map (applySignature graph i_0) signatures

     -- any simple signatures reached the final interval?
     satisfies is = not (and (map (isEmpty . (i_f /\)) is))
     reachable_seq = map satisfies iss
     satisfying0 = zip signatures reachable_seq 
     
     -- any loop_end signature reached the final interval?
     satisfying1 = 
      [ (initSignature signature !!++ toSignature 
            [initPath (lastSignature signature), loop, toPath [headPath loop]]
        , test graph loop i_f intervals)
      | (signature, intervals) <- zip signatures iss
      , loop <- loops
      ]

     satisfying = satisfying0 ++ satisfying1


-- Compute the fixpoints of a given MAF
fixpoints :: MAF -> Interval RealPlus -> Interval RealPlus -> Interval RealPlus
fixpoints maf interval rnge =
  let (af1, af2) = fromMAF maf

      (l, u) = fromInterval interval
      (l', u') = (value l, value u)
      
      (rl, ru) = fromInterval rnge
      rnge' = (value rl, value ru)
  in toInterval (open $ fixpoints' af1 l' rnge', open $ fixpoints' af2 u' rnge')


fixpoints' af x0 (ex1, ex2)
   | a < 1  = fixpoints_lt_1 
   | a > 1  = fixpoints_gt_1 
   | a == 1 = fixpoints_eq_1
   where
     (a, b) = fromAF af
     unstablefp = b/(1-a)

     fixpoints_eq_1 
       | b == 0 && x0 == ex1 = InfiNeg
       | b == 0 && x0 == ex2 = InfiPos
       | b < 0               = InfiNeg
       | b == 0              = x0
       | b > 0               = InfiPos

     fixpoints_lt_1
       | x0 &== unstablefp && x0 == ex1 = InfiNeg
       | x0 &== unstablefp && x0 == ex2 = InfiPos
       | otherwise                      = unstablefp

     fixpoints_gt_1
       | unstablefp &== ex1 && ex1==x0 = InfiPos
       | unstablefp &== ex2 && ex2==x0 = InfiNeg
       | x0 &== unstablefp             = x0
       | x0 &< unstablefp              = InfiNeg
       | x0 &> unstablefp              = InfiPos

x &== y = x-y <= epsilon && y-x <= epsilon
x &< y = x < y+epsilon
x &> y = x+epsilon > y
epsilon = 0.0000001

-- Given an TMAF and a new edge transition to follow, return the new 
-- TMAF (composing the transition at the end)
oneStepComp :: Graph -> TMAF -> (EdgeId,EdgeId) -> TMAF 
--oneStepComp graph tmaf0 (src_id, dst_id) = 
--  comp tmaf0 (fromJust $ graphTMAF graph src_id dst_id)
oneStepComp graph tmaf0 (src_id, dst_id) = 
  let tmaf1 = graphTMAF graph src_id dst_id
      errorMessage = "It is not feasible to go from Edge "++ show src_id ++ " to Edge " ++ show dst_id
  in  case tmaf1 of
        Nothing -> error errorMessage 
        Just f  -> comp tmaf0 (fromJust tmaf1)


-- Calculate the whole TMAF for a (sequential) path
multiStepComp :: Graph -> Path -> TMAF
multiStepComp graph path = foldl (oneStepComp graph) funcId pairs
  where
    pairs = [ (x,y) | x:y:_ <- tails (fromPath path) ]



loopType :: Graph -> Path -> Maybe LoopType
loopType graph loop 
     | isEmpty sIntersectLoop             = Nothing -- "Loop cannot be taken"
     | (extr_L <= fpl) && (fpu <= extr_U) = Just $ Stay (fpl, fpu)
     | (fpl > extr_U) || (fpu < extr_L)   = Just $ Die 
     | (fpl < extr_L) && (fpu > extr_U)   = Just $ ExitBoth 
     | (extr_L <= fpl) && (extr_U < fpu)  = Just $ ExitRight fpl
     | (fpl < extr_L) && (fpu <= extr_U)  = Just $ ExitLeft fpu
    where
      tmaf0 = multiStepComp graph loop
      last_s = lastPath loop

      loop_tmaf = oneStepComp graph tmaf0 (last_s, headPath loop)

      (s,j) = fromTMAF2SJ loop_tmaf
      int_LU = s /\ j

      (extr_L, extr_U) = values int_LU

      (extr_l, extr_u) = fromInterval (fixpoints (fromTMAF2MAF loop_tmaf) int_LU (edgeRange graph $ headPath loop))
      (fpl, fpu) = (value extr_l, value extr_u)

      appliedLoop = apply loop_tmaf int_LU
      sIntersectLoop = appliedLoop /\ s -- gives the iterable interval: i \in s/\j

