-- User Input Library for the SPDI application
-- Gordon Pace - VERIMAG 
-- 26/07/2001

module UserInput (
  module Input,
  module Vector, 
  module AffineFunctions,
 
  UserInputRegion(..), userRegionVertices, userRegionVFirst, userRegionVSecond, 
  UserInput (..), vertexTable, userRegions, pointNames,

  getParitiesOfRegion, getParities, userInputToInput
) where

import Maybe
import List

import Vector
import AffineFunctions
import Input
import Interval
import Signature

type VertexId = Int 

data UserInputRegion =
  UserInputRegion {
    userRegionVertices :: [VertexId],
    userRegionVFirst, userRegionVSecond :: Vector
  } deriving (Eq, Show)
  
data UserInput =
  UserInput {
    pointNames  :: [String],
    vertexTable :: [Vector], 
    userRegions :: [UserInputRegion]
  } deriving (Eq, Show)


getParities :: UserInput -> [[Parity]]
getParities userinput = map (getParitiesOfRegion (vertexTable userinput)) (userRegions userinput)
    
getParitiesOfRegion :: [Vector] -> UserInputRegion -> [Parity]
getParitiesOfRegion vertex_table region = parities
  where
    region_vector = userRegionVFirst region
    region_vertices = map getCoordinates (userRegionVertices region)
    rotation = head
      [ (v3-v1) `isClockwiseOf` (v2-v1) 
      | v1:v2:v3:_ <- tails (region_vertices++take 2 region_vertices), not(areColinear v1 v2 v3)
      ]
    parities = [ if region_vector `isClockwiseOf` (v2-v1) == rotation then In else Out
               | v1:v2:_ <- tails (region_vertices ++ [head region_vertices])
               ]

    getCoordinates n = vertex_table !! n

------------------
-- Transformation of the User input into the internal representation 
-- of the input
userInputToInput :: UserInput -> 
  (Input, (String, String) -> [EdgeId], EdgeId -> (String, String))
userInputToInput userinput = 
  (Input { 
    inputEdges   = zip [0..] (map normalize_edge edge_table), 
    inputRegions = regions'
  }, points2edgeID, edgeID2points )
  where
    -- calculate fetchedge
    labels = pointNames userinput

    edgeID2points eid = (labels !! v1, labels !! v2)
      where
        ((v1,v2),_) = edge_table !! eid

    points2edgeID (p1, p2) 
      | isNothing id1 || isNothing id2 = []
      | otherwise = result
      where
        id1 = findIndex (p1 ==) labels
        id2 = findIndex (p2 ==) labels
        (id1',id2') = (fromJust id1, fromJust id2)
        table = zip [0..] $ map fst edge_table
        result = [ n | (n,(id1, id2)) <- table, id1==id1', id2==id2' ] ++ 
                 [ n | (n,(id1, id2)) <- table, id1==id2', id2==id1' ]
                     
    -- calculate actual input
    vertexPositions = vertexTable userinput
    regions = userRegions userinput

    -- Which edges do we need?
    (edge_table, regions') = 
        mapAccumL addEdgesInRegion initialEdges regions

    -- Edge information
    initialEdges = [] -- [((VertexId,VertexId), RealPlus)]

    edgeId edge_table (extremes, len) =
      case findIndex ((extremes ==) . fst) edge_table of
        Nothing -> (edge_table ++ [(extremes, len)], length edge_table)
        Just i  -> (edge_table, i)

    -- handling one region
    addEdgesInRegion knownEdges region =
      (knownEdges', 
       Region {
         vFirst   = v_first,
         vSecond  = v_second,
         edgeInfo = zip edge_ids parities
       }
      )
      where
        -- Retreive information
        parities = getParitiesOfRegion vertexPositions region
        vertices = userRegionVertices region

        v_first  = userRegionVFirst region
        v_second = userRegionVSecond region

        -- Is the region closed?
        closed = head vertices == last vertices

        -- Which edges need their direction flipped?
        flip = 
          let p1:p2:_ = vertices
              noflip  = (vertex p2 - vertex p1) >.< hat v_first > 0
              parity:_ = parities
          in if noflip then opposite parity else parity

        -- The edges we need, and their lengths
        edges0 = [(src,dst) | src:dst:_ <- tails vertices]
        lengths0 = map (\(s,d) -> modulus $ vertex d - vertex s) edges0

        edges1 = zipWith 
            (\(s,d) parity -> if parity==flip then (d,s) else (s,d))
            edges0 parities
        edges_length = 
          if closed 
            then zip edges1 lengths0 
            else fixLengths0 edges1 lengths0 parities
  
        fixLengths0 (e:es) (l:ls) (p:ps) = 
           (e, if p==flip then InfiPos else InfiNeg):fixLengths es ls ps
        fixLengths [e] _ [p] = if p==flip then [(e,InfiNeg)] else [(e,InfiPos)]
        fixLengths (e:es) (l:ls) (p:ps) = (e, l):fixLengths es ls ps
 
        -- collect the new edge information
        (knownEdges', edge_ids) = mapAccumL edgeId knownEdges edges_length


    -- normalize edges
    normalize_edge ((src, dst), len) = 
       let src' = vertex src
           dst' = vertex dst 
           vec  = dst' - src'
       in Edge { 
            position = if len==InfiNeg then dst' else src', 
            unitDir  = vec / constant' (modulus vec),
            lambdaRange = 
              case len of
                InfiPos -> toInterval (closed 0, closed InfiPos)
                InfiNeg -> toInterval (closed InfiNeg, closed 0)
                _       -> toInterval (closed 0, closed len)
          }
    vertex n = vertexPositions !! n 


