-- Vector Library
-- Gordon Pace - Gerardo Schneider - VERIMAG 
-- 16/07/2001

module Vector (
  module Infinity,
  Vector(..), fromVector,
  constant, constant', reflect, (>.<), hat, modulus, determinant, distance,
  rotateVector, vectorToAngle, angleToVector, sumVectors, 
  isClockwiseOf, isAnticlockwiseOf, isClockwiseOfOrOn, isAnticlockwiseOfOrOn,
  areColinear, readVectors, 

  Line(..), Location(..), intersectLines, isOnLine, onLine
) where

import Useful
import Infinity

newtype Vector = Vector (RealPlus, RealPlus) deriving Eq
newtype Line  = Line  (Vector, Vector) deriving Eq

vector (x,y) = Vector (Finite x, Finite y)

fromVector (Vector p) = p
unLine  (Line  l) = l

instance Num Vector where
-- Sum of vectors
  Vector (x1,y1) + Vector (x2,y2) = Vector (x1+x2, y1+y2)
-- Difference of vectors
  Vector (x1,y1) - Vector (x2,y2) = Vector (x1-x2, y1-y2)
-- Product of vectors
  Vector (x1,y1) * Vector (x2,y2) = Vector (x1*x2, y1*y2)
-- Negation of vectors
  negate (Vector (x,y)) = Vector (-x, -y)

  abs (Vector (x,y)) = Vector (abs x, abs y)
  signum (Vector (x,y)) = Vector (signum x, signum y)

  fromInteger n = let r = fromInteger n in Vector (r,r)

sumVectors :: [Vector] -> Vector
sumVectors = foldl (+) 0

instance Fractional Vector where
  Vector (x1,y1) / Vector (x2,y2) = Vector (x1/x2, y1/y2)
  
  fromRational r = let r' = fromRational r in Vector (r',r')

-- constant vector
constant :: Double -> Vector
constant x = let x' = Finite x in Vector (x',x')

constant' :: RealPlus -> Vector
constant' x = Vector (x,x)

-- reflect along x=y
reflect :: Vector -> Vector
reflect (Vector (x,y)) = Vector (y,x)

-- dot (inner) product
(>.<) :: Vector -> Vector -> RealPlus
p1 >.< p2 = let (x,y) = fromVector (p1 * p2) in x+y

-- vector rotation
rotateVector :: Double -> Vector -> Vector
rotateVector angle v = 
  let (x,y) = fromVector v 
      alpha = angle * pi / 180 
      cosA  = Finite $ cos alpha
      sinA  = Finite $ sin alpha
  in Vector (x * cosA - y * sinA, y * cosA + x * sinA)

-- angle subtended by the vector with the x-axis
vectorToAngle :: Vector -> Double
vectorToAngle p 
  | x==0      = if y>0 then 90 else -90
  | otherwise = fix ((180*atan(toDouble (y / x)))/ pi)
  where
    (x,y) = fromVector p
    fix angle = if x<0 then angle+180 else angle

-- a unit vector in the direction of the requested angle
angleToVector :: Double -> Vector
angleToVector theta
  | abs (theta-90) <= epsilon = vector (1,0)
  | abs (theta+90) <= epsilon = vector (-1,0)
  | theta < -180 = angleToVector (theta + 360)
  | theta >  180 = angleToVector (theta - 360)
  | otherwise    = v / constant' (modulus v)
  where
    epsilon = 0.000001
    v = fix $ Vector (1.0, Finite (tan (theta * pi/ 180)))
    fix vector = if abs theta > 90 then Vector (-1,-1)*vector else vector

-- are three vectors colinear?
areColinear v1 v2 v3 = (v1-v3) >.< (v2-v3) == 0

-- is v3 clockwise wrt to v2 (origin v1)
v1 `isClockwiseOf` v2 = v1 >.< hat v2 > 0
v1 `isAnticlockwiseOf` v2 = v2 `isClockwiseOf` v1
v1 `isClockwiseOfOrOn` v2 = v1 >.< hat v2 >= 0
v1 `isAnticlockwiseOfOrOn` v2 = v2 `isClockwiseOfOrOn` v1

-- Orthogonal vector: (x, y) |-> (y, -x)
hat :: Vector -> Vector
hat (Vector (x,y)) = Vector (y, -x)

-- Norm
modulus vector = sqrtPlus (vector >.< vector)

-- Distance
distance v1 v2 = modulus (v1 - v2)

-- determinant
determinant :: Vector -> Vector -> RealPlus
determinant p1 p2 = let (x,y) = fromVector (p1 * reflect p2) in x - y

instance Show Vector where
  show = show . fromVector

instance Show Line where
  show (Line (p1, p2)) = 
    "The line going from "++show p1++" to "++show p2

data Location = 
  BeforeStart | OnLine | BeyondEnd
  deriving (Eq, Show)

isOnLine OnLine = True
isOnLine _ = False

-- Line gradients
type Gradient = RealPlus

gradient :: Vector -> Vector -> Gradient
gradient p1 p2 = 
  let (dx, dy) = fromVector (p2 - p1) in
    if dx == 0 then InfiPos else dy / dx

-- precondition for onLine is that the points are colinear
onLine :: Vector -> Line -> Bool
p `onLine` (Line (q,q')) =
  (x_p `within` (x,x')) && 
  (y_p `within` (y,y'))
  where
    (x_p, y_p) = fromVector p
    (x, y)  = fromVector q
    (x',y') = fromVector q'

    n `within` (m,m') = n >= min m m' && n <= max m m
       
{--
Taken from the comp.graphics.algorithms FAQ:

Subject 1.03: How do I find intersections of 2 2D line segments?

    This problem can be extremely easy or extremely difficult depends
    on your applications.  If all you want is the intersection point,
    the following should work:

    Let A,B,C,D be 2-space position vectors.  Then the directed line
    segments AB & CD are given by:

        AB=A+r(B-A), r in [0,1]
        CD=C+s(D-C), s in [0,1]

    If AB & CD intersect, then

        A+r(B-A)=C+s(D-C), or

        Ax+r(Bx-Ax)=Cx+s(Dx-Cx)
        Ay+r(By-Ay)=Cy+s(Dy-Cy)  for some r,s in [0,1]

    Solving the above for r and s yields

            (Ay-Cy)(Dx-Cx)-(Ax-Cx)(Dy-Cy)
        r = -----------------------------  (eqn 1)
            (Bx-Ax)(Dy-Cy)-(By-Ay)(Dx-Cx)

            (Ay-Cy)(Bx-Ax)-(Ax-Cx)(By-Ay)
        s = -----------------------------  (eqn 2)
            (Bx-Ax)(Dy-Cy)-(By-Ay)(Dx-Cx)

    Let P be the position vector of the intersection point, then

        P=A+r(B-A) or

        Px=Ax+r(Bx-Ax)
        Py=Ay+r(By-Ay)

    By examining the values of r & s, you can also determine some
    other limiting conditions:

        If 0<=r<=1 & 0<=s<=1, intersection exists
            r<0 or r>1 or s<0 or s>1 line segments do not intersect

        If the denominator in eqn 1 is zero, AB & CD are parallel
        If the numerator in eqn 1 is also zero, AB & CD are collinear.

    If they are collinear, then the segments may be projected to the x- 
    or y-axis, and overlap of the projected intervals checked.

    If the intersection point of the 2 lines are needed (lines in this
    context mean infinite lines) regardless whether the two line
    segments intersect, then

        If r>1, P is located on extension of AB
        If r<0, P is located on extension of BA
        If s>1, P is located on extension of CD
        If s<0, P is located on extension of DC

    Also note that the denominators of eqn 1 & 2 are identical.

    References: [O'Rourke (C)] pp. 249-51 [Gems III] pp. 199-202 
    "Faster Line Segment Intersection,"
--}

intersectLines :: Line -> Line -> Maybe (Vector, Location, Location)
intersectLines line1 line2 = 
  let
    (a,b) = unLine line1
    (c,d) = unLine line2

    a_c = a - c
    d_c = d - c
    b_a = b - a

    denominator = determinant b_a d_c 

    r = determinant d_c a_c / denominator
    s = determinant b_a a_c / denominator 
 
    vector = a + constant' r * b_a in

    if denominator == 0 -- parallel
       then Nothing
       else Just (vector, check r, check s)
  
    where
      check p
        | p < 0     = BeforeStart
        | p > 1     = BeyondEnd
        | otherwise = OnLine

-- This code is to hide away translation of lists of pairs or lists of lists of
-- pairs of strings into the equivalent constructs but of vectors.

class ReadVectors a b where
  readVectors :: Monad m => (String -> Maybe Vector) -> a -> m b

instance ReadVectors String Vector where
  readVectors point2vector pointname =
    case point2vector pointname of
      Nothing    -> fail ("Point <"++pointname++"> undefined")
      Just point -> return point

instance (ReadVectors a x, ReadVectors b y) => ReadVectors (a,b) (x,y) where
  readVectors userinput (x,y) =
    do { x' <- readVectors userinput x; y' <- readVectors userinput y; return (x',y') }

instance ReadVectors a b => ReadVectors [a] [b] where
  readVectors userinput [] = return []
  readVectors userinput (p:ps) =
    do { p' <- readVectors userinput p; ps' <- readVectors userinput ps; return (p':ps') }


